#include <math.h>
#include "geo_distance.h"
#include "scia_common.h"

/* Formulas for distance calculations  */
double PI_2 = 1.5707963268;
double DEG2RAD = 0.0174532925;
/* 				define acos */
double acos2(double x)  
{ 
  if (x == 1.0)
    return 0.0;
  if (x == -1.0)
    return 2.0*PI_2;
  return PI_2 - atan2(x, sqrt(1-x*x));
}

/*  define length formula */
/*  based on formula for shperical trigonometrie (Bronstein, p. 208) */
/*  triangle (lat0,lon0), north pole, (lat1,lon1) */

double deltal3 (double lat0, double lat1, double dlon)
{
  double a = PI_2 - lat1;
  double b = PI_2 - lat0;
  return acos2( cos(a)*cos(b) + sin(a)*sin(b)*cos(dlon));
}


double circle_distance ( Coord_deg p0, Coord_deg p1 )
{    
  return deltal3( p0.lat * DEG2RAD, p1.lat * DEG2RAD,
		  (p0.lon-p1.lon) * DEG2RAD );
}



/*  function inpoly : check for point inside or outside polygon */

/*    1=inside, 0=outside                   */
/*  is target point inside a 2D polygon?  */
/*    polygon points, [0]=x, [1]=y        */
/*    number of points in polygon         */
/*    x (horizontal) of target point      */
/*    y (vertical) of target point        */
/*  local variables */

int  inpoly ( double *polyx, double *polyy, int npoints,
	      double xt, double yt)
{
    double x1, x2, y1, y2, xnew, xold, ynew, yold;
    int i,inside;
    
    inside=0;
				/*  need 3 points to be inside */
				/*  nothing is inside one line */
    if (npoints < 3) {
	return(0);
    }
				/*  This is an optimized crossing boundary
				    check */
    xold=polyx[npoints-1];
    yold=polyy[npoints-1];    
    for (i=0 ; i < npoints ; i++) {
	xnew=polyx[i];
	ynew=polyy[i];
	if (xnew > xold) {
	    x1=xold;
	    x2=xnew;
	    y1=yold;
	    y2=ynew;
	}
	else {
	    x1=xnew;
	    x2=xold;
	    y1=ynew;
	    y2=yold;
	}
	if ((xnew < xt) == (xt <= xold) && (yt-y1)*(x2-x1) < (y2-y1)*(xt-x1))
	{			/* sum up boundary crossings */
	    inside++;
	}
	xold=xnew;
	yold=ynew;
    }
				/* return result */
    return(inside%2);
}


#define MAX_POLY 10

/*  do the inside check for geographical areas (not to large) */
/*  here the 0/360 boundary is checked */
/*    1=inside, 0=outside                   */

int check_inside (Coord_deg point, Coord_deg* area, int n_coords)
{
    int n_neg, n_pos, i;

    double plat, plon;
    double lat[MAX_POLY];
    double lon[MAX_POLY];
    
    n_neg=0;
    n_pos=0;

    plat = point.lat;
    plon = point.lon;	
			     /*  handle this awfull 180/-180 boundary */
                             /*  by putting the complete rectangle to the */
                             /*  180 or -180 line			 */
    for (i=0; i<n_coords; i++)
    {
	lon[i] = area[i].lon;	/* Copy to temp. arrays */
	lat[i] = area[i].lat;
				/*  */
	if (area[i].lon < -90.0)  n_neg++;
	if (area[i].lon > 90.0)   n_pos++;
    }
				/*  Check if corner points at both sides
				    of the date boundary */
    if (n_neg > 0 && n_pos > 0)
    {
				/*  yes? than map all points to the side of
				    the search point */
	if (plon < -90)
	{
	    for (i=0; i<n_coords; i++)
	    {
		if (lon[i] > 90)
		    lon[i] = lon[i] - 360;
	    }
	}
	else
	{
	    for (i=0; i<n_coords; i++)
	    {
		if (lon[i] < -90)
		    lon[i] = lon[i] + 360;
	    }
	}
    }

				/*  now we can perform the check */
    return inpoly(lon,lat,n_coords, plon, plat);
    
}

/* Call inpoly function with corners as given in geolocation */

int point_in_gp (Coord_deg *corners, Coord_deg point)
{
    Coord_deg area[4];
    area[0] = corners[0];
    area[1] = corners[1];
    area[2] = corners[3];
    area[3] = corners[2];
    return check_inside (point, area, 4);
}
    



