/* #include "scia_defs.h" */
/* #include "scia_common.h" */
#include "scia_L2_lib.h"
/* #include "bin_read.h" */
#include "geo_distance.h"


/***********************************************************************
Description of Limb/Occultation Time calculations (IODD 0-1, p 57)
------------------------------------------------------------------------
    The measurement duration of a Limb state is 59.0625 seconds. This
    may be calculated as follows: one azimuth scan has a net
    integration time of 1.5 seconds and the adjustment of the
    elevation mirror for the next azimuth scan takes 3 BCPSs
    corresponding to 0.1875 seconds giving a total of 1.6875 seconds
    for one azimuth scan. 34 azimuth scan plus one dark measurement
    into deep space (which takes the same time as one azimuth scan)
    gives a state duration of: 35 x 1.6875 s = 59.0625 s.
**********************************************************************/
/* change in 2003 via OCR to improve limb-nadir matching:
   further on only 30 azimuth scan plus one dark measurement,
    gives a state duration of: 31 x 1.6875 s = 52.3125 s (837BCPS)
                duration_scan_state = 836 BCPS
**************************************************************/
SCIA_err openL2N (char* FILE_name, info_l2N *info)

{
    int n,i,err,n_geo;
    Gome_Doas gome;
    Gome_Doas_4 gome_4;

    info->FILE_l2N = fopen(FILE_name, "rb");
    
    if (info->FILE_l2N == NULL) {
	fprintf(stderr,"Input file %s could not be opened! Abort!\n", FILE_name);
	return FILE_NOT_FOUND;
    }

    if ( gome_check (info->FILE_l2N) )
    {
	info->scia_product = GOME_L2;
	MPH_GOME_getbin (info->FILE_l2N, &info->mph_gome);
    }
    else
    {

	/*  MPH, SPH, DSDs */
	if (Read_MPH (info->FILE_l2N, &info->mph) != OK)
	    return MPH_ERROR;
    
	info->scia_product = scia_product (&info->mph);
	if (info->scia_product != SCIA_L2N &&
	    info->scia_product != SCIA_L2O &&
	    info->scia_product != SCIA_L2M &&
	    info->scia_product != MIPAS_L2)
	    return MPH_ERROR;	/* invalid product */
	
	strncpy (info->orbit_date, info->mph.sensing_start, MAX_UTC_STRING);

    }

    /*  set some pointers to zero */
    info->states_L = NULL;
    info->geo = NULL;
    info->geo_L = NULL;
    info->data_str = NULL;


    switch (info->scia_product)
    {
	case SCIA_L2N:
	    if (Read_SPH_SCI_NL__2P (info->FILE_l2N, &info->sph) != OK)
		return SPH_ERROR;
	    for ( err=0,i=0; i<SCI_NL__2P_MAX_DS_NAME; i++)
		err += Read_DSD ( info->FILE_l2N, &info->dsd[i] );
	    if (err != 0)
		return DSD_ERROR;
	    /*  Read Geolocation */
	    n_geo = info->dsd[ SCI_NL__2P_GEOLOCATION ].num_dsr;
	    if (n_geo > 0)
	    {
		if (set_DSD_offset( info->FILE_l2N,
				    &info->dsd[ SCI_NL__2P_GEOLOCATION ] ) != OK)
		    return ERROR;

		info->geo = malloc ( sizeof (GeoN_L2_user) * n_geo );
		GeoN_L2_user_array_getbin (info->FILE_l2N, info->geo, n_geo);
	    }
	    info->n_geo = n_geo;
	    /* Initialize info */
	    for (i=0; i< SCI_NL__2P_MAX_DS_NAME; i++)
	    {
		info->cur_offset[i] = info->dsd[i].offset;
		info->cur_n[i] = 0;
		info->geo_n[i] = 0;
	    }
	    break;

	case SCIA_L2M:
	    if (Read_SPH_SCI_RV__2P (info->FILE_l2N, &info->sph_met) != OK)
		return SPH_ERROR;
	    for ( err=0,i=0; i < SCI_RV__2P_MAX_DS_NAME; i++)
		err += Read_DSD ( info->FILE_l2N, &info->dsd[i] );
	    if (err != 0)
		return DSD_ERROR;
	    /*  Read Geolocation */
	    n_geo = info->dsd[ SCI_RV__2P_GEOLOCATION ].num_dsr;
	    if (n_geo > 0)
	    {
		if (set_DSD_offset( info->FILE_l2N,
				    &info->dsd[ SCI_RV__2P_GEOLOCATION ] ) != OK)
		    return ERROR;

		info->geo = malloc ( sizeof (GeoN_L2_user) * n_geo );
		GeoN_L2_user_met_array_getbin (info->FILE_l2N, info->geo, n_geo);
	    }
	    info->n_geo = n_geo;
		
	    for (i=0; i< SCI_RV__2P_MAX_DS_NAME; i++)
	    {
		info->cur_offset[i] = info->dsd[i].offset;
		info->cur_n[i] = 0;
		info->geo_n[i] = 0;
	    }
	    break;

	case SCIA_L2O:
	    if (Read_SPH_SCI_OL__2P (info->FILE_l2N, &info->sph_ol) != OK)
		return SPH_ERROR;
	    for ( err=0,i=0; i < SCI_OL__2P_MAX_DS_NAME; i++)
		err += Read_DSD ( info->FILE_l2N, &info->dsd[i] );
	    if (err != 0)
		return DSD_ERROR;
	    /*  Read Geolocation NADIR */
	    n_geo = info->dsd[ SCI_OL__2P_GEOLOCATION_NADIR ].num_dsr;
	    if (n_geo > 0)
	    {
		if (set_DSD_offset( info->FILE_l2N,
				    &info->dsd[ SCI_OL__2P_GEOLOCATION_NADIR ] )
		    != OK)
		    return ERROR;

		info->geo = malloc ( sizeof (GeoN_L2_user) * n_geo );
		GeoN_L2_OL_user_array_getbin (info->FILE_l2N, info->geo, n_geo);
	    }
	    info->n_geo = n_geo;
	    /*  Read Geolocation LIMB */
	    n_geo = info->dsd[ SCI_OL__2P_GEOLOCATION_LIMB ].num_dsr;
	    if (n_geo > 0)
	    {
		if (set_DSD_offset( info->FILE_l2N,
				    &info->dsd[ SCI_OL__2P_GEOLOCATION_LIMB ] )
		    != OK)
		    return ERROR;

		info->geo_L = malloc ( sizeof (GeoL_L2_user) * n_geo );
		GeoL_L2_OL_user_array_getbin (info->FILE_l2N,
					      info->geo_L, n_geo);
	    }
	    info->n_geo_L = n_geo;
	    
	    for (i=0; i< SCI_OL__2P_MAX_DS_NAME; i++)
	    {
		info->cur_offset[i] = info->dsd[i].offset;
		info->cur_n[i] = 0;
		info->geo_n[i] = 0;
		/*  initialise counters to zero */
		info->n_geoprofile_in_state[i] = 0;
		info->n_geoprofile_per_state[i] = 0;
		info->states_n[i] = 0;
	    }
	    /*  States info needed for unique time_stamp */
	    /*  of Limb profiles, not used for nadir */
	    info->n_states = info->dsd[ SCI_OL__2P_STATES ].num_dsr;
	    if (info->n_states > 0)
	    {
		if (set_DSD_offset( info->FILE_l2N,
				    &info->dsd[ SCI_OL__2P_STATES ] )
		    != OK)
		    return ERROR;
		info->states_L = malloc ( sizeof (States) * info->n_states);
		States_array_getbin ( info->FILE_l2N,
				      info->states_L, info->n_states);
	    }
	    
	    break;

	case GOME_L2:
	    SPH_GOME_getbin (info->FILE_l2N, &info->sph_gome);
	    /* Number of fitting windows */
	    info->gome_nwin = info->sph_gome.Nwin;
	    /* Number of molecules */
	    info->gome_nmol = info->sph_gome.Nmol;
	    /* Set version switch (GDP < 4 / GDP >= 4) */
	    if (strncmp (info->sph_gome.version_DOAS, "04.00", 5)
		< 0)
		info->gome_ver = GOME;
	    else
		info->gome_ver = GOME_4;	      
	    /* Evaluate products available in product */
	    for (i=0; i<GOME_L2_MAX; i++)
		/* preset -1  used as not available  */
		info->gome_l2_mol_idx[i] = -1;

	    for (n=0; n<info->gome_nmol; n++)
	    {
		for (i=0; i<GOME_L2_MAX; i++)
		{
		    if (strncmp (GOME_L2_MOL_NAME[i],
				 info->sph_gome.molpair[n],
				 strlen (GOME_L2_MOL_NAME[i])) == 0)
			break;
		}
		if (i==GOME_L2_MAX)
		{
		    fprintf (stderr, "WARNING: Products not implemented here\n"
			     "found in GOME product:  *** %s ***\n"
			     "This product cannot accessed with this version of %s.!\n"
			     "Update  list in gome_L2.c/gome_L2.h :\n"
			     " GOME_L2_MOL_NAME / GOME_L2_MOL_WIN / GOME_L2_MAX.\n");
		    /* Not a fatal error, a warning is enough */
		    /* return SCIA_GOME_EXTRA_PRODUCT; */
		}
		info->gome_l2_mol_idx[i] = n;
		info->gome_l2_win_idx[i] = GOME_L2_MOL_WIN[i];
	    }
		
	    /* read first pixel to get start time of orbit */
	    i = (int) ftell (info->FILE_l2N);
	    
	    switch (info->gome_ver)
	    {
		case GOME:
		    Gome_Doas_getbin (info->FILE_l2N,
				      &gome,
				      info->gome_nmol,
				      info->gome_nwin);
		break;
		case GOME_4:
		    Gome_Doas_4_getbin (info->FILE_l2N,
					&gome_4,
					info->gome_nmol,
					info->gome_nwin);
		    break;

		default:
		    return ERROR;
		    
	    }	    
	    /* go back before first pixel */
	    if (fseek(info->FILE_l2N, i, SEEK_SET) == -1)
	    {
		return ERROR;
	    }

	    UTC_String (&gome.geo.dsr_time, info->orbit_date);


	    for (i=0; i< SCI_OL__2P_MAX_DS_NAME; i++)
	    {
		info->cur_n[i] = 0;
	    }

	    break;
	    
	case MIPAS_L2:
	    if (Read_SPH_MIP_NL__2P (info->FILE_l2N, &info->sph_mipas) != OK)
		return SPH_ERROR;
	    for ( err=0,i=0; i <  MAX_DS_NAME_MIP_NL__2P; i++)
		err += Read_DSD ( info->FILE_l2N, &info->dsd[i] );
	    if (err != 0)
		return DSD_ERROR;
	    /*  extract number and order of species */
	    info->N_gas = 0;
	    /*  This is NRT product */
	    if (strcmp (info->sph_mipas.order_of_species,
			"H2O,O3,HNO3,CH4,N2O,NO2       " ) == 0)
	    {
		info->N_gas = 6;
		info->species[0] = MIPAS_H2O;
		info->species[1] = MIPAS_O3;
		info->species[2] = MIPAS_HNO3;
		info->species[3] = MIPAS_CH4;
		info->species[4] = MIPAS_N2O;
		info->species[5] = MIPAS_NO2;
	    }
    
	    if (strcmp (info->sph_mipas.order_of_species,
			"O3,H2O                        " ) == 0)
	    {
		info->N_gas = 2;
		info->species[1] = MIPAS_H2O;
		info->species[0] = MIPAS_O3;
	    }

	    if (info->N_gas == 0)
	    {
		fprintf (stderr, "MIPAS: Order of product has changed according to SPH!\n"
			 "programm has to be changed!\n");
		return SPH_ERROR;
	    }		
	    /*  Read Dataset_structure */
	    info->n_data_str = info->dsd[ DATASET_STRUCTURE_ADS ].num_dsr;
	    if (set_DSD_offset( info->FILE_l2N,
				&info->dsd[ DATASET_STRUCTURE_ADS ] ) != OK)
		return ERROR;

	    info->data_str = malloc ( sizeof (Dataset_Structure_Ads) *
				      info->n_data_str );

	    for (i=0; i< info->n_data_str; i++)
		Dataset_Structure_Ads_getbin (info->FILE_l2N, &info->data_str[i]);
	    /*  set all counters to initial values */
	    for (i=0; i< MAX_DS_NAME_MIP_NL__2P; i++)
	    {
		info->cur_offset[i] = info->dsd[i].offset;
		info->cur_n[i] = 0;
		info->geo_n[i] = 0;
	    }
	    info->cur_data_str = 0;
	    info->N_swp = 
		info->data_str[info->cur_data_str].num_sweeps;
	    break;

	default:
	    return MPH_ERROR;
    }	    	    
    return OK;
}

SCIA_err closeL2N (info_l2N *info)
{
    fclose (info->FILE_l2N);
    if (info->geo)
	free (info->geo);
    if (info->geo_L)
	free (info->geo_L);
    if (info->states_L)
	free (info->states_L);
    if (info->data_str)
        free (info->data_str);
    return OK;
}

/* perform seek to */
/* the current position */
SCIA_err MDS_seek (info_l2N *info, int prod)
{
    if ( fseek (info->FILE_l2N, info->cur_offset[prod],
		SEEK_SET) == -1 )
    {
	return ERROR;
    }
    return OK;
}



/* Main routine of this package: Reads the MDS one by one! */
/* Also calls the geolocation-coadding routines */

SCIA_err L2N_next_data (info_l2N *info, product_id prod,
			user_data *ud)
{
    BIAS_L2N bias;		/* L2N bias */
    DOAS_L2N doas;		/* L2N doax */
    Nadir_OL nad_ol;		/* L2OL nadir (doas+bias) */
    O3_met met;			/* L2met */
				/* Aerosol parts */
    CLOUDS_AEROSOL_L2N cld_aer;
    Clouds_Aerosol_L2O cld_aer_ol;
    /*  L2OL limb */
    Limb_OL lim_ol;
    /*  GOME */
    Gome_Doas gome;
    Gome_Doas_4 gome_4;
  
      
    MJD *mds_time;
    /*  temporary geolocation */
    GeoN_L2_user geo;
    GeoL_L2_user geo_L;

    /*  actual number of levels in limb profiles */
    int n_level;
    /*  index of (first) geolocation of scans
	in limb profiles*/
    int index_geo_L [MAX_LIMB_LEVELS];
  
    int integr_time;
    int geo_coadd;
  
    char fitwin_id_str[2];
    int fitwin_id;
    unsigned n;
    int i,m;

    int n_prod;			/* internal product_id for easy handling */

    char gome_orbit[6];
    /* needed to switch between OL data version (add. diagnostics)*/
    float ol_ver=0.0;

    /*  */
    /*  Determine the needed method */
    /*  */
    switch (info->scia_product)
    {
	/* MIPAS product */
	case MIPAS_L2:
	    ud->method = MIPAS;
	    n_prod = prod.prod_mip;
	    break;
	/*  METEO product */
	case SCIA_L2M:
	    ud->method = METEO;
	    if (prod.prod_m != SCI_RV__2P_O3_VERTICAL_COLUMN)
		/*  invalid product */
		return SCIA_BAD_SPECIFICATION;
	    n_prod = prod.prod_m;
	    break;
	    /*  NRT product */
	case SCIA_L2N:
	    switch (prod.prod_n)
	    {
		case SCI_NL__2P_CLOUDS_AEROSOL:
		    ud->method = CLD_AER;
		    break;
		case   SCI_NL__2P_DOAS_0_O3:
		case   SCI_NL__2P_DOAS_1_NO2:
		case   SCI_NL__2P_DOAS_1_H2O:
		case   SCI_NL__2P_DOAS_1_O3:
		case   SCI_NL__2P_DOAS_2_BRO:
		case   SCI_NL__2P_DOAS_2_O3_L:
		case   SCI_NL__2P_DOAS_2_O3_H:
		case   SCI_NL__2P_DOAS_2_NO2:
		case   SCI_NL__2P_DOAS_2_OCLO:
		case   SCI_NL__2P_DOAS_3_OCLO:
		case   SCI_NL__2P_DOAS_3_NO2:
		case   SCI_NL__2P_DOAS_3_O4:
		case   SCI_NL__2P_DOAS_4_SO2:
		case   SCI_NL__2P_DOAS_4_O3:
		case   SCI_NL__2P_DOAS_5_HCHO:
		case   SCI_NL__2P_DOAS_5_BRO:
		case   SCI_NL__2P_DOAS_5_O3_L:
		case   SCI_NL__2P_DOAS_5_O3_H:
		case   SCI_NL__2P_DOAS_5_NO2:
		case   SCI_NL__2P_DOAS_5_O4:
		case   SCI_NL__2P_DOAS_SPARE_1:
		case   SCI_NL__2P_DOAS_SPARE_2:
		    ud->method = DOAS;
		    break;
		case   SCI_NL__2P_BIAS_0_H2O:
		case   SCI_NL__2P_BIAS_0_CO2:
		case   SCI_NL__2P_BIAS_1_N2O:
		case   SCI_NL__2P_BIAS_1_H2O:
		case   SCI_NL__2P_BIAS_1_CH4:
		case   SCI_NL__2P_BIAS_2_CO:
		case   SCI_NL__2P_BIAS_2_H2O:
		case   SCI_NL__2P_BIAS_2_CH4:
		case   SCI_NL__2P_BIAS_SPARE_1:
		case   SCI_NL__2P_BIAS_SPARE_2:
		    ud->method = BIAS;
		    break;
		default:
		    /*  invalid product */
		    return SCIA_BAD_SPECIFICATION;
		    break;
	    }
	    n_prod = prod.prod_n;
	    break;
	    /*  OL product */
	case SCIA_L2O:
	    /*  check flag for limb (from user option) */
	    if (info->limb_flag == NADIR)
	    {
		switch (prod.prod_o)
		{
		    case  SCI_OL__2P_CLOUDS_AEROSOL:
			ud->method = CLD_AER_OL;
			break;
		    case  SCI_OL__2P_NAD_UV0_O3:
		    case  SCI_OL__2P_NAD_UV1_NO2:
		    case  SCI_OL__2P_NAD_UV2_O3:
		    case  SCI_OL__2P_NAD_UV3_BRO:
		    case  SCI_OL__2P_NAD_UV4_H2CO:
		    case  SCI_OL__2P_NAD_UV5_SO2:
		    case  SCI_OL__2P_NAD_UV6_OCLO:
		    case  SCI_OL__2P_NAD_UV7_SPARE:
		    case  SCI_OL__2P_NAD_IR0_H2O:
		    case  SCI_OL__2P_NAD_IR1_CH4:
		    case  SCI_OL__2P_NAD_IR2_N2O:
		    case  SCI_OL__2P_NAD_IR3_CO:
		    case  SCI_OL__2P_NAD_IR4_CO2:
		    case  SCI_OL__2P_NAD_IR5_SPARE:
			ud->method = NADIR_OL;
			break;
		    default:
			return SCIA_BAD_SPECIFICATION;
			break;
		}
		n_prod = prod.prod_o;
	    }
	    else
	    {
		switch (prod.prod_l)
		{	
		    case SCI_OL__2P_LIM_PTH:		      
		    case SCI_OL__2P_LIM_UV0_O3:
		    case SCI_OL__2P_LIM_UV1_NO2:
		    case SCI_OL__2P_LIM_UV2_O3:
		    case SCI_OL__2P_LIM_UV3_BRO:
		    case SCI_OL__2P_LIM_UV4_H2CO:
		    case SCI_OL__2P_LIM_UV5_SO2:
		    case SCI_OL__2P_LIM_UV6_OCLO:
		    case SCI_OL__2P_LIM_UV7_SPARE:
		    case SCI_OL__2P_LIM_IR0_H2O:
		    case SCI_OL__2P_LIM_IR1_CH4:
		    case SCI_OL__2P_LIM_IR2_N2O:
		    case SCI_OL__2P_LIM_IR3_CO:
		    case SCI_OL__2P_LIM_IR4_SPARE:
			ud->method = LIMB_OL;
			break;

			/*  invalid product */
		    default:
			return SCIA_BAD_SPECIFICATION;
			break;		  
		}
		n_prod = prod.prod_l;
	    }
	    break;

	    /*  GOME */
	case GOME_L2:
	    /* GOME or GOME_4 */
	    ud->method = info->gome_ver;
	    n_prod = prod.prod_g;
	    break;
	  
	default:			/* Not a level-2 product !!! */
	    return SCIA_BAD_SPECIFICATION;
	    break;
    }

    /*  Do some checks : requested product available? */
    switch (ud->method)
    {				/* SCIA */
	case METEO:
	case BIAS:
	case DOAS:
	case CLD_AER:
	case NADIR_OL:
	case CLD_AER_OL:
	case LIMB_OL:	  
	    /*  this products not available? */
	    if (info->dsd[n_prod].num_dsr == 0)
		return SCIA_MDS_EMPTY;
	    /*  All MDS already read? */
	    if (info->cur_n[n_prod] >= info->dsd[n_prod].num_dsr)
		return SCIA_END_OF_MDS;
	    if ( MDS_seek (info, n_prod) != OK)
		return ERROR;
	    break;
				/* GOME */
	case GOME:
	case GOME_4:
	    if (info->gome_l2_mol_idx [n_prod] == -1)
		return SCIA_MDS_EMPTY;
	    if (info->cur_n[n_prod] >= info->mph_gome.n_doas)
		return SCIA_END_OF_MDS;
	    break;
	case MIPAS:
	    if (info->cur_n[SCAN_INFORMATION_MDS] >=
		info->dsd[SCAN_INFORMATION_MDS].num_dsr)
		return SCIA_END_OF_MDS;
	    if ( MDS_seek (info, SCAN_INFORMATION_MDS) != OK)
		return ERROR;
	    break;
    }
	    
    /*  */
    /*  Now read the mds, according to method !!*/
    /*  */
    switch (ud->method)
    {
	case METEO:
	    O3_met_getbin (info->FILE_l2N, &met);
	    mds_time = &met.dsr_time;
	    integr_time =  met.integr_time;
	    break;

	case BIAS:
	    /* read MDS */
	    BIAS_L2N_getbin (info->FILE_l2N, &bias);
	    mds_time = &bias.dsr_time;
	    integr_time =  bias.integr_time;
	    break;
	case DOAS:
	    DOAS_L2N_getbin (info->FILE_l2N, &doas);
	    mds_time = &doas.dsr_time;
	    integr_time = doas.integr_time;
	    break;
	case CLD_AER:
	    CLOUDS_AEROSOL_L2N_getbin (info->FILE_l2N, &cld_aer);
	    mds_time = &cld_aer.dsr_time;
	    integr_time = cld_aer.integr_time;
	    break;
	case NADIR_OL:
	    /* read MDS */
	    Nadir_OL_getbin (info->FILE_l2N, &nad_ol);
	    mds_time = &nad_ol.dsr_time;
	    integr_time = nad_ol.integr_time;
	    break;
	case CLD_AER_OL:
	    Clouds_Aerosol_L2O_getbin (info->FILE_l2N, &cld_aer_ol);
	    mds_time = &cld_aer_ol.dsr_time;
	    integr_time = cld_aer_ol.integr_time;
	    break;
	case LIMB_OL:	  
	    Limb_OL_getbin (info->FILE_l2N, &lim_ol);
	    mds_time = &lim_ol.dsr_time;
	    integr_time = lim_ol.integr_time;	/* not used in limb */
	    /* In test product, there are empty MDS in product, to be ignored */
	    /* 1: empty  0: contain data */
	    if (lim_ol.n_main == 0 || lim_ol.n_meas == 0)
	    {
		info->cur_n[n_prod]++;
		return SCIA_LIMB_MDS_EMPTY;
	    }
	    /* evaluate add. diagnostics if version > 3.0 */
	    sscanf (info->mph.software_ver + 8, "%f", &ol_ver);
	    if (ol_ver >= 3.0)
	    {
		int n;
		int n_diag = 0;
		int n_main = lim_ol.n_main;
		int n1 = lim_ol.n1;
		int n_state_vec = lim_ol.n_state_vec;
		lim_ol.add_diagnostic.degree_of_freedom = lim_ol.add_diag[n_diag++];
		lim_ol.add_diagnostic.info_content = lim_ol.add_diag[n_diag++];
		for (n=0; n<n_state_vec; n++)
		    lim_ol.add_diagnostic.diag_covar_matrix[n] = lim_ol.add_diag[n_diag++];
		for (n=0; n< n_main*n1; n++)
		    lim_ol.add_diagnostic.num_den[n] = lim_ol.add_diag[n_diag++];
		for (n=0; n< n_main*n1; n++)
		    lim_ol.add_diagnostic.ini_num_den[n] = lim_ol.add_diag[n_diag++];
		for (n=0; n< n_main; n++)
		    lim_ol.add_diagnostic.conv_vmr[n] = lim_ol.add_diag[n_diag++];
		for (n=0; n< n_main; n++)
		    lim_ol.add_diagnostic.conv_num_den[n] = lim_ol.add_diag[n_diag++];
		for (n=0; n<n_state_vec; n++)
		    lim_ol.add_diagnostic.aver_kernel[n] = lim_ol.add_diag[n_diag++];
		    
		if (n_diag > lim_ol.n_ad)
		{
		    fprintf (stderr, "L2N_next_data: n_diag > lim_ol.ad : %d > %d\n", 
			     n_diag, lim_ol.n_ad);
		    return SCIA_PRODUCT_INCONSISTENCY;
		}
		
	    }
	    /* set this stuff to zero for version < 3.0 */
	    else
	    {
		int n;
		int n_diag = 0;
		int n_main = lim_ol.n_main;
		int n1 = lim_ol.n1;
		int n_state_vec = lim_ol.n_state_vec;
		lim_ol.add_diagnostic.degree_of_freedom = 0.0;
		lim_ol.add_diagnostic.info_content = 0.0;
		for (n=0; n<n_state_vec; n++)
		    lim_ol.add_diagnostic.diag_covar_matrix[n] = 0.0;
		for (n=0; n< n_main*n1; n++)
		    lim_ol.add_diagnostic.num_den[n] =  0.0;
		for (n=0; n< n_main*n1; n++)
		    lim_ol.add_diagnostic.ini_num_den[n] =  0.0;
		for (n=0; n< n_main; n++)
		    lim_ol.add_diagnostic.conv_vmr[n] =  0.0;		
		for (n=0; n< n_main; n++)
		    lim_ol.add_diagnostic.conv_num_den[n] =  0.0;
		for (n=0; n<n_state_vec; n++)
		    lim_ol.add_diagnostic.aver_kernel[n] =  0.0;		    
	    }		
	    break;
	case GOME:
	    Gome_Doas_getbin (info->FILE_l2N,
			      &gome,
			      info->gome_nmol,
			      info->gome_nwin);
	    break;
	case GOME_4:
	    Gome_Doas_4_getbin (info->FILE_l2N,
			      &gome_4,
			      info->gome_nmol,
			      info->gome_nwin);
	    break;
	case MIPAS:
	    Scan_Information_Mds_getbin (info->FILE_l2N, &ud->scan, info->N_swp,
					 info->N_gas);
	    info->cur_offset[SCAN_INFORMATION_MDS] = (int) ftell ( info->FILE_l2N);

	    break;
	    
    }

    if (ud->method != GOME && ud->method != GOME_4 && ud->method != MIPAS)
    {
	/*  update offset (SCIA) */
	info->cur_offset[n_prod] = (int) ftell ( info->FILE_l2N);
    }
    /*  ********************************    */
    /* Now find the appropriate geolocation */
    /*  *********************************   */
    switch (ud->method)
    {				/* Nadir geolocation */
	case METEO:
	case BIAS:
	case DOAS:
	case CLD_AER:
	case NADIR_OL:
	case CLD_AER_OL:
      
	    if (MJD_compare (mds_time, &info->geo[ info->geo_n[n_prod] ].dsr_time)
		!= 0)
	    {
		for ( n = info->geo_n[n_prod]+1; n < info->n_geo; n++ )
		{
		    if (MJD_compare (mds_time, &info->geo[ n ].dsr_time)
			== 0)
			break;
		}
		/*  Paranoia check : Geolocation available? */
		if ( n == info->n_geo)
		    {printf ("-->%2d\n",1); return SCIA_GEO_MISMATCH;}
		info->geo_n[n_prod] = n;
	    }
	    /* add up geo */
	    GeoN_L2_add (&geo, &geo_coadd, info, n_prod, integr_time);

	    /* build user_data - geo - nadir */
	    ud->geo = geo;
	    UTC_String (&geo.dsr_time, ud->date_str);
	    ud->frac_day = JDAY_01012000 (&geo.dsr_time);

	    break;
	    /*  ***************** */
	    /*  Limb geolocation  */
	    /*  ***************** */
	case LIMB_OL:
	    /*  do we need to search for new state info?*/
    NEXT_PROFILE:
	    if (info->n_geoprofile_in_state[n_prod] == 0)
	    {
		States *cur_state = NULL;
		/*  search for corresponding state */
		/*  to find state start time */
		/*  for unique limb time stamp */
		for (n = info->states_n[n_prod]; n < info->n_states; n++)
		{
		    double tdiff;
		    tdiff = MJD_diff (&info->states_L[n].dsr_time, mds_time);
				/* Measurement before available state */
		    if (tdiff < 0 ) 
			{printf ("-->%2d\n",2); return SCIA_GEO_MISMATCH;}
				/*  Measurements not in this state, 
				    skip  and goto next */
		    if (tdiff > info->states_L[n].duration_scan_state
			/ BCPS_FACTOR)
			continue;
				/*  here we found our state */
				/*  store it */
		    UTC_String (&info->states_L[n].dsr_time,
				ud->state_date_str);
		    cur_state = info->states_L + n;
		    ud->state = info->states_L + n;
		    ud->n_state = n;
		    info->states_n[n_prod] = n;		    
		    break;
		}
		/*  No corresponding state found ? */
		if ( n == info->n_states)
		{
		    return SCIA_GEO_MISMATCH;
		}
		  	      
		/*  calculate number of geolocation profiles
		    from integration time */
		info->n_geoprofile_per_state[n_prod] =
		    BCPS_PER_LIMBSCAN / cur_state->shortest_int_time;
		/* number of levels from state duration */
		/*  duration_scan_state+1 / 27 */
		info->n_levels_in_state[n_prod] = 
		    (cur_state->duration_scan_state + 1 )
		    / BCPS_LIMBSCAN_TO_LIMBSCAN;
		/*  search start index in geolocation for
		    this state, start at alredy used geo_locations */
		for ( n = info->geo_n[n_prod]; n < info->n_geo_L; n++ )
		{
		    if (MJD_compare (&cur_state->dsr_time,
				     &info->geo_L[ n ].dsr_time) == 0)
			break;
		}
		/*  Paranoia check : Geolocation available? */
		if ( n == info->n_geo_L)
		    {
			return SCIA_GEO_MISMATCH;
		    }
		/*  store start index of geolocation */
		info->geo_n[n_prod] = n;

		/* uncomment for debugging purposes... */
		/*
		{
		    char tmp_date[28];
		    char tmp_date1[28];

		UTC_String (&cur_state->dsr_time, tmp_date);

		fprintf (stderr, "\nState %2d  %2d  %s\n",
			info->states_n[n_prod],
			info->n_geoprofile_per_state[n_prod], tmp_date);

		for (i=0; i<info->n_geoprofile_per_state[n_prod]*35; i++)
		{
		    double tdiff = 0.0;
		    if (i>1)
			tdiff = MJD_diff (&info->geo_L[info->geo_n[n_prod]+ i - 1 ].dsr_time,
					  &info->geo_L[info->geo_n[n_prod]+ i ].dsr_time);

		    UTC_String (&info->geo_L[info->geo_n[n_prod]+ i ].dsr_time, tmp_date);
		    fprintf (stderr," %3d %s %7.4f\n", i, tmp_date,tdiff);
		}
		} */

	    } /* state setup is finished */

	    /* Now for eache profile: */

	    /*  set up index list to geolocation of
		profile, recognizing scan pattern:
		0 1 2 3
		7 6 5 4
	    */	
	    /* set local n_level from previous stored state info */
	    n_level = info->n_levels_in_state[n_prod];
	    /* check it */
	    if (n_level > 35 || n_level < 0)
	    {
		fprintf (stderr, "Fatal error: More than 35 levels in "
			 "Limb sequence detected! %d", n_level);
		exit (255);
	    }

	    for ( n=0; n<n_level; n++)
	    {
		switch (n%2)
		{
		    case 0:
			index_geo_L[n] = info->geo_n[n_prod] +
			    info->n_geoprofile_in_state[n_prod] +
			    n * info->n_geoprofile_per_state[n_prod];
			break;
		    case 1:
			index_geo_L[n] = info->geo_n[n_prod] -
			    info->n_geoprofile_in_state[n_prod] - 1 +
			    (n+1) * info->n_geoprofile_per_state[n_prod];
			break;
		}
/*		fprintf (stderr, "index_geo_L[%2d] : %4d\n",
			 n,index_geo_L[n]);
*/
	    }

	    /*  This should never happen */
	    if (lim_ol.n_meas > n_level)
		return SCIA_GEO_MISMATCH;
	    /*  Number of measurements is number of
		geolocations */
	    ud->n_geo_L = lim_ol.n_meas;
	    /*  Search for start index of geolocation,
		using last dsr_time of 
		measurement_grid[ ].dsr_time ( because of
		top down retrieval) and
		index_geo_L[n] */
	    for ( n=0; n<n_level; n++)
	    {
		int cmp;
		cmp = MJD_compare
		    (&lim_ol.measurement_grid[lim_ol.n_meas-1].dsr_time,
		     &info->geo_L[ index_geo_L[n] ].dsr_time);
		/* uncomment this for debug purposes */
		/*
		{
		    char tmp_date[28];
		    char tmp_date1[28];

		    UTC_String (&info->geo_L[ index_geo_L[n] ].dsr_time,
				tmp_date);
		    UTC_String (&lim_ol.measurement_grid[lim_ol.n_meas-1].dsr_time,
				tmp_date1);
		    fprintf (stderr, "%2d %4d %s %s %d\n", n, index_geo_L[n], 
			     tmp_date, tmp_date1, cmp);
		}
		*/
		if (cmp == 0)
		{
		    ud->n_geo_0 = n;
		    break;
		}
		

	    }

	    if ( n == n_level )
	    {
		/* a profile is missing */
		/* enlarge profile_in_state counter modulo maximum number */
		/*  and simply jump back to try next profile */
		info->n_geoprofile_in_state[n_prod] =
		    (info->n_geoprofile_in_state[n_prod] + 1) %
		    info->n_geoprofile_per_state[n_prod];
		goto NEXT_PROFILE;
	    }
	    /*  make paranoia check for correct
		geolocation */

	    for ( n = 0; n < lim_ol.n_meas; n++ )
	    {
		int i = lim_ol.n_meas-n-1;
		/* i index to used measurement */
		m = ud->n_geo_0 + n;
		/* m index to corresponding geoloaction
		 */

		/* time of both should be equal */
	        if (MJD_compare
		    (&lim_ol.measurement_grid[i].dsr_time,
		     &info->geo_L[ index_geo_L[m] ].dsr_time) != 0)
		    {
			;
			/*fprintf ( stderr, "Warning: SCIA_GEO_MISMATCH ignored\n");*/
		    /* Error for many profiles, not that critical */
		    /* give a warning and continue... */
/*		    return SCIA_GEO_MISMATCH;} */
		    }

	    }
	    /* store number of levels in user_data */
	    ud->n_level = n_level;
  
	    /*  Now set up list of geolocation for
		user_data doing co-add if necessary!
	    */
	    geo_coadd = lim_ol.integr_time /
		info->states_L[info->states_n[n_prod]].shortest_int_time;
	  
	    for ( n = 0; n<n_level; n++ )
	    {
		Geo_limb_coadd (info,
				&ud->geo_L_array[n_level - 1 - n],
				&info->geo_L[ index_geo_L[n] ],
				geo_coadd);
		/*  because we have in Limb alternating
		    scan directions, it is user friendly
		    to swap every second scan start and end of
		    scan for angles */
		if ((n_level - 1 - n)%2 == 1)
		    Geo_limb_swap (&ud->geo_L_array[n_level - 1 - n]);
		/* 
		fprintf (stderr,
			 " ud->geo_L_array[%d].tangent_height[1] : %f\n",
			 n_level - 1 - n,
			 ud->geo_L_array[n_level - 1 - n].tangent_height[1]); 
		*/
	    }

	  
	    /* Calculate mean geolocation  */
	    Geo_Limb_mean (ud->geo_L_array, &geo_L , n_level);
	    /*  build corner coordinates */
	    ud->corner_limb[0] = ud->geo_L_array[n_level-2].tangent_coord[0];
	    ud->corner_limb[1] = ud->geo_L_array[n_level-2].tangent_coord[2];
	    ud->corner_limb[2] = ud->geo_L_array[0].tangent_coord[2];
	    ud->corner_limb[3] = ud->geo_L_array[0].tangent_coord[0];
	    /*  Profile identifier */
	    ud->subsetcounter = info->n_geoprofile_in_state[n_prod] *
		ud->geo_L.integr_time;
	    ud->n_profile_in_state = info->n_geoprofile_in_state[n_prod];
	    /* geo_coadd used  only in nadir */
	    /* set to zero to avoid possiblse
	       side effects */
	    geo_coadd = 0;
	    /* build user_data - geo - limb */
	    ud->geo_L = geo_L;
	    UTC_String (&geo_L.dsr_time, ud->date_str);
	    ud->frac_day = JDAY_01012000 (&geo_L.dsr_time);	  	      
	    break;

	    /*  GOME geolocation*/
	case GOME:
	    ud->geo = gome.geo;
	    UTC_String (&ud->geo.dsr_time, ud->date_str);
	    ud->frac_day = JDAY_01012000 (&ud->geo.dsr_time);
	    break; 
	case GOME_4:
	    ud->geo = gome_4.geo;
	    UTC_String (&ud->geo.dsr_time, ud->date_str);
	    ud->frac_day = JDAY_01012000 (&ud->geo.dsr_time);
	    break; 
	case MIPAS:
  				/*  Geolocation */
	    if (info->cur_n[SCAN_GEOLOCATION_ADS] >=
		info->dsd[SCAN_GEOLOCATION_ADS].num_dsr)
		return SCIA_END_OF_MDS;
	    if ( fseek (info->FILE_l2N, info->cur_offset[SCAN_GEOLOCATION_ADS],
			SEEK_SET) == -1 )
	    {
		return ERROR;
	    }
	    Scan_Geolocation_Ads_getbin (info->FILE_l2N, &ud->geo_mipas);
	    info->cur_offset[SCAN_GEOLOCATION_ADS] = (int) ftell ( info->FILE_l2N);
				/*  check, if next DATASET_STRUCTURE starts */
	    if (info->cur_data_str < info->dsd[DATASET_STRUCTURE_ADS].num_dsr-1)
		if (MJD_compare (&ud->geo_mipas.dsr_time,
				 &info->data_str[info->cur_data_str+1].dsr_time) == 0)
		{
		    info->cur_data_str ++;
		    info->N_swp = 
			info->data_str[info->cur_data_str].num_sweeps;
		}
				/*  This is not a rectangle, just set to something */
				/*  to avoid wrong output */
///	    ud->corner_limb[0] = ud->goe_mipas.geo_loc_first;
//	    ud->corner_limb[1] = ud->goe_mipas.geo_loc_first;
//	    ud->corner_limb[2] = ud->goe_mipas.geo_loc_last;
//	    ud->corner_limb[3] = ud->goe_mipas.geo_loc_last;

	    UTC_String (&ud->geo_mipas.dsr_time, ud->date_str);
	    ud->frac_day = JDAY_01012000 (&ud->geo_mipas.dsr_time);
	    ud->N_swp = info->N_swp;

	    break;
	  
    }
	  
    /* set the general (static) information in userdata ud */
    switch (ud->method)
    {      
	case GOME:
	case GOME_4:
	    ud->product = info->mph_gome.pir;
	    ud->software_ver = info->sph_gome.version_DOAS;
	    strncpy (gome_orbit, info->mph_gome.pir+5,5);
	    gome_orbit[5]='\0';
	    ud->abs_orbit = atoi (gome_orbit);
	    ud->prod = prod;
	    ud->orbit_date = info->orbit_date; 
	    break;

	default:			/* SCIA */
	    /* build user_data */
	    ud->product = info->mph.product;
	    ud->software_ver = info->mph.software_ver;
	    ud->abs_orbit = atoi (info->mph.abs_orbit);
	    ud->prod = prod;
	    ud->orbit_date = info->orbit_date; 
	    break;
    }

    /* set molecule name and fitwindow */
    switch (ud->method)
    {
	case METEO:
	    /*  Use Name from NRT (its the same data!)*/
	    ud->molecule = DS_NAME_SCI_NL__2P[SCI_NL__2P_DOAS_0_O3]+5;
	    ud->doas_fitting_window = info->sph_met.doas_fitting_window_0;      
	    ud->met = met;
	    break;
	case BIAS:
	    ud->molecule = DS_NAME_SCI_NL__2P[n_prod]+5;
	    fitwin_id_str[0] = DS_NAME_SCI_NL__2P[n_prod][5];
	    fitwin_id_str[1] = '\0';
	    fitwin_id = atoi (fitwin_id_str);
	    ud->bias_fitting_window = info->sph.bias_fitting_window[ fitwin_id ];
	    ud->bias = bias;
	    break;
	case DOAS:
	    ud->molecule = DS_NAME_SCI_NL__2P[n_prod]+5;
	    fitwin_id_str[0] = DS_NAME_SCI_NL__2P[n_prod][5];
	    fitwin_id_str[1] = '\0';
	    fitwin_id = atoi (fitwin_id_str);
	    ud->doas_fitting_window = info->sph.doas_fitting_window[ fitwin_id ];
	    ud->doas = doas;
	    break;
	case CLD_AER:
	    ud->molecule = DS_NAME_SCI_NL__2P[n_prod];
	    ud->cld_aer = cld_aer;
	    /* for cld_aer, this should always be equal */
	    if (info->geo_n[n_prod] != info->cur_n[n_prod])
	    {
		{printf ("-->%2d\n",6); return SCIA_GEO_MISMATCH;}
	    }
	    break;
	case NADIR_OL:
	    ud->molecule = DS_NAME_SCI_OL__2P[n_prod]+6;
	    fitwin_id_str[0] = DS_NAME_SCI_OL__2P[n_prod][6];
	    fitwin_id_str[1] = '\0';
	    fitwin_id = atoi (fitwin_id_str);
	    if (n_prod < SCI_OL__2P_NAD_IR0_H2O) /* UV prod */
	    {
		ud->doas_fitting_window = info->sph_ol.nad_fit_window_uv[ fitwin_id ];
		ud->n_prod = 1;
		ud->main_prod = 0;
	    }
	    else
	    {
		int n;
		char tmp[20];
		ud->bias_fitting_window = info->sph_ol.nad_fit_window_ir[ fitwin_id ];
		/* extract fitted molecule and default molecule number */
		ud->n_prod = sscanf (ud->bias_fitting_window,
				     " %s %s %s %s %s %s %s ",
				     tmp,
				     ud->prod_str_ir [0],
				     ud->prod_str_ir [1],
				     ud->prod_str_ir [2],
				     ud->prod_str_ir [3],
				     ud->prod_str_ir [4],
				     ud->prod_str_ir [5]) - 1;
		for (n=0; n<ud->n_prod; n++)
		{
		    if (strcmp (ud->molecule+2,  ud->prod_str_ir[n]) == 0)
		    {
			ud->main_prod = n;
			break;
		    }
		}
		if (ud->n_prod < 2 || n == ud->n_prod)
		    return ERROR;
	    }
	    ud->nad_ol = nad_ol;
	    break;
	case CLD_AER_OL:
	    ud->molecule = DS_NAME_SCI_NL__2P[n_prod];
	    ud->cld_aer_ol = cld_aer_ol;
	    /* for cld_aer, this should always be equal */
	    if (info->geo_n[n_prod] != info->cur_n[n_prod])
	    {
		{printf ("-->%2d\n",7); return SCIA_GEO_MISMATCH;}
	    }
	  
	case LIMB_OL:
	    /* no fitting window for PTH ?? */
	    if (n_prod == SCI_OL__2P_LIM_PTH)
		break;
	    ud->molecule = DS_NAME_SCI_OL__2P[n_prod]+6;
	    fitwin_id_str[0] = DS_NAME_SCI_OL__2P[n_prod][6];
	    fitwin_id_str[1] = '\0';
	    fitwin_id = atoi (fitwin_id_str);
	    if (n_prod < SCI_OL__2P_LIM_IR0_H2O) /* UV prod */
		ud->doas_fitting_window = info->sph_ol.lim_fit_window_uv[ fitwin_id ];
	    else
		ud->bias_fitting_window = info->sph_ol.lim_fit_window_ir[ fitwin_id ];	      
	    ud->lim_ol = lim_ol;
	    break;

	case GOME:
	    ud->gome = gome;
	    ud->molecule = GOME_L2_MOL_NAME[n_prod];
				/*  For GOME , main_prod is definded as
				 index of the requested species */
	    ud->main_prod = info->gome_l2_mol_idx [ n_prod ];
				/*  Some infos are stored according to the
				    fit window */
	    ud->main_win = info->gome_l2_win_idx [ n_prod ];
	    break;

	case GOME_4:
	    ud->gome_4 = gome_4;
	    ud->molecule = GOME_L2_MOL_NAME[n_prod];

				/*  For GOME , main_prod is definded as
				 index of the requested species */
	    ud->main_prod = info->gome_l2_mol_idx [ n_prod ];
				/*  Some infos are stored according to the
				    fit window */
	    ud->main_win = info->gome_l2_win_idx [ n_prod ];
	    break;
	    
	case MIPAS:
	    ud->species_str = mipas_species[n_prod];
	    ud->pos_cur_species = n_prod;
	    break;
    }

    /* Update counters */
    switch (ud->method)
    {				/* update Nadir counters */
	case METEO:
	case BIAS:
	case DOAS:
	case CLD_AER:
	case NADIR_OL:
	case CLD_AER_OL:
	    info->geo_n[n_prod] += geo_coadd;
	    info->cur_n[n_prod]++;
	    break;
	    /*  update Limb counters */
	case LIMB_OL:
	    info->n_geoprofile_in_state[n_prod] =
		(info->n_geoprofile_in_state[n_prod] + 1) %
		info->n_geoprofile_per_state[n_prod];
	    info->cur_n[n_prod]++;
	    break;
	    /* update GOME counters */
	case GOME:
	case GOME_4:
	    info->cur_n[n_prod]++;
	    break;
	case MIPAS:
	    info->cur_n[SCAN_GEOLOCATION_ADS]++;
	    break;
    }      
    return OK;
}

	      




SCIA_err GeoN_L2_add (GeoN_L2_user *geo_coadd, int *n_coadd, info_l2N *info,
		      int prod, int integr_time)
{
    GeoN_L2_user *geo = info->geo;
    int geo_n = info->geo_n[prod];

    int i,n0,n1,n2,nd;
    /* Whatever times involved, this is the same */
    *n_coadd = integr_time / geo[geo_n].integr_time;
    /*  paranoia check */
    if (*n_coadd * geo[geo_n].integr_time != integr_time)
	return SCIA_PRODUCT_INCONSISTENCY;
    /*  Start time is from first pixel*/
    geo_coadd->dsr_time = geo[geo_n].dsr_time;
    /*  Attach flag from 1. !!!!!!!!*/
    geo_coadd->attach_flag = geo[geo_n].attach_flag;
    /*  Integration time is calculated */
    geo_coadd->integr_time = integr_time;
    /*  Mean of sat_geod_ht and earth_rad */
    /*  over this time period */
    geo_coadd->sat_geod_ht = 0;
    geo_coadd->earth_rad = 0;	
    for (i=0; i< *n_coadd; i++)
    {
	geo_coadd->sat_geod_ht += geo[geo_n+i].sat_geod_ht;
	geo_coadd->earth_rad += geo[geo_n+i].earth_rad;
    }
    geo_coadd->sat_geod_ht /= *n_coadd;
    geo_coadd->earth_rad /= *n_coadd;

    /* All possibilities less than 16 (1s) */
    /* No complications with integration of backscan, at least
       5 readouts for complete scan sequence */
    if (integr_time <= 16)
    {
	/*  subset_counter is from first pixel */
	geo_coadd->subset_counter = geo[geo_n].subset_counter ;

	/* Set angles and coordinates */
	switch (*n_coadd)
	{
	    case 1:
		/*  angles*/
		for (i=0; i<3; i++)
		{
		    geo_coadd->sol_zen_angle_toa[i] =
			geo[geo_n].sol_zen_angle_toa[i];
		    geo_coadd->los_zen_angle_toa[i] =
			geo[geo_n].los_zen_angle_toa[i];
		    geo_coadd->rel_azi_angle_toa[i] =
			geo[geo_n].rel_azi_angle_toa[i];
		}
		/*  coordinates */
		geo_coadd->cen_coor_nad = geo[geo_n].cen_coor_nad;
		for (i=0; i<4; i++)
		    geo_coadd->cor_coor_nad[i] = geo[geo_n].cor_coor_nad[i];
		break;
		/*  coadd two geolocations */
		/*  old: 0--1--2/0--1--2   */
		/*  new: 0------1------2   */
	    case 2:
	    case 4:
	    case 8:
	    case 16:
		n0 = geo_n;
		n1 = geo_n + *n_coadd/2 -1;
		n2 = geo_n + *n_coadd -1;
		/*  angles */
		geo_coadd->sol_zen_angle_toa[0] =
		    geo[n0].sol_zen_angle_toa[0];
		geo_coadd->sol_zen_angle_toa[1] =
		    geo[n1].sol_zen_angle_toa[2];
		geo_coadd->sol_zen_angle_toa[2] =
		    geo[n2].sol_zen_angle_toa[2];
		
		geo_coadd->los_zen_angle_toa[0] =
		    geo[n0].los_zen_angle_toa[0];
		geo_coadd->los_zen_angle_toa[1] =
		    geo[n1].los_zen_angle_toa[2];
		geo_coadd->los_zen_angle_toa[2] =
		    geo[n2].los_zen_angle_toa[2];
		
		geo_coadd->rel_azi_angle_toa[0] =
		    geo[n0].rel_azi_angle_toa[0];
		geo_coadd->rel_azi_angle_toa[1] =
		    geo[n1].rel_azi_angle_toa[2];
		geo_coadd->rel_azi_angle_toa[2] =
		    geo[n2].rel_azi_angle_toa[2];
		/*  coordinates */
		/*  should be replaced by a more
		    sophisticated approach */
		geo_coadd->cen_coor_nad =
		    middle_coord (geo[n1].cen_coor_nad,
				  geo[n1+1].cen_coor_nad);
		geo_coadd->sub_sat_point =
		    middle_coord (geo[n1].sub_sat_point,
				  geo[n1+1].sub_sat_point);
		/*  corners */
		geo_coadd->cor_coor_nad[0] = geo[n0].cor_coor_nad[0];
		geo_coadd->cor_coor_nad[1] = geo[n0].cor_coor_nad[1];
		geo_coadd->cor_coor_nad[2] = geo[n2].cor_coor_nad[2];
		geo_coadd->cor_coor_nad[3] = geo[n2].cor_coor_nad[3];
		break;
	    default:
		return SCIA_PRODUCT_INCONSISTENCY;
		break;
	}
	return OK;
    }
    /* Integratiot times larger the 16 (1s) */
    if (integr_time == 160)
    {
	/*  subset_counter here is -2 */
	geo_coadd->subset_counter = -2;
	switch (*n_coadd)
	{
	    case 10:
	    case 20:
	    case 40:
	    case 80:
	    case 160:
		n0 = geo_n;	/*  */
		nd = *n_coadd / 10; /* n_delta */
		n1 = geo_n + nd*5 ; 
		/* angles */
		geo_coadd->sol_zen_angle_toa[0] =
		    (geo[ n0 ].sol_zen_angle_toa[0]
		     + geo[ n1 ].sol_zen_angle_toa[0]) / 2;
		geo_coadd->sol_zen_angle_toa[1] =
		    (geo[ n0 + 2*nd - 1 ].sol_zen_angle_toa[2]
		     + geo[ n1 + 2*nd - 1 ].sol_zen_angle_toa[2]) / 2;
		geo_coadd->sol_zen_angle_toa[2] =
		    (geo[ n0 + 4*nd-1 ].sol_zen_angle_toa[2]
		     + geo[ n1 + 4*nd-1 ].sol_zen_angle_toa[2]) / 2;

		geo_coadd->los_zen_angle_toa[0] =
		    (geo[ n0 ].los_zen_angle_toa[0]
		     + geo[ n1 ].los_zen_angle_toa[0]) / 2;
		geo_coadd->los_zen_angle_toa[1] =
		    (geo[ n0 + 2*nd - 1 ].los_zen_angle_toa[2]
		     + geo[ n1 + 2*nd - 1 ].los_zen_angle_toa[2]) / 2;
		geo_coadd->los_zen_angle_toa[2] =
		    (geo[ n0 + 4*nd-1 ].los_zen_angle_toa[2]
		     + geo[ n1 + 4*nd-1 ].los_zen_angle_toa[2]) / 2;
	
		geo_coadd->rel_azi_angle_toa[0] =
		    (geo[ n0 ].rel_azi_angle_toa[0]
		     + geo[ n1 ].rel_azi_angle_toa[0]) / 2;
		geo_coadd->rel_azi_angle_toa[1] =
		    (geo[ n0 + 2*nd - 1 ].rel_azi_angle_toa[2]
		     + geo[ n1 + 2*nd - 1 ].rel_azi_angle_toa[2]) / 2;
		geo_coadd->rel_azi_angle_toa[2] =
		    (geo[ n0 + 4*nd-1 ].rel_azi_angle_toa[2]
		     + geo[ n1 + 4*nd-1 ].rel_azi_angle_toa[2]) / 2;
		/*  coordinates */
		/*  should be replaced by a more
		    sophisticated approach */
		geo_coadd->cen_coor_nad = geo[ n0 + 2*nd - 1 ].cor_coor_nad[3];
		/* sub-satellite point: mean point between first and last */
		geo_coadd->sub_sat_point = middle_coord
		    (geo[n0].sub_sat_point,
		     geo[n1].sub_sat_point);
		
		/*  corners */
		geo_coadd->cor_coor_nad[0] = geo[n0].cor_coor_nad[0];
		geo_coadd->cor_coor_nad[1] = geo[n1].cor_coor_nad[1];
		geo_coadd->cor_coor_nad[2] = geo[ n0 + 4*nd-1 ].cor_coor_nad[2];
		geo_coadd->cor_coor_nad[3] = geo[ n1 + 4*nd-1 ].cor_coor_nad[3];
		break;
	    case 1:
		return SCIA_GEO_INTEGR_TIME_NOT_IMPLEMENTED;
		break;
	    default:
		return SCIA_PRODUCT_INCONSISTENCY;
		break;
	}
	return OK;
    }
    return SCIA_GEO_INTEGR_TIME_NOT_IMPLEMENTED;
}


/* Combine subsequent Limb geolocation to one geolocation */

SCIA_err Geo_limb_coadd (info_l2N *info,
			 GeoL_L2_user *geo_coadd,
			 GeoL_L2_user *geo_L,
			 int n_coadd)
{
    int n0,n1,n2,i;
    if (n_coadd == 1)		/* nothing to do */
    {
	*geo_coadd = *geo_L;
	return OK;
    }
    /*  Start time is from first pixel*/
    geo_coadd->dsr_time = geo_L->dsr_time;
    /*  Attach flag from 1. !!!!!!!!*/
    geo_coadd->attach_flag = geo_L->attach_flag;
    /*  New Integration time  */
    geo_coadd->integr_time = geo_L->integr_time * n_coadd;
    /*  Mean of sat_geod_ht and earth_rad */
    /*  over this time period */
    geo_coadd->sat_geod_ht = 0;
    geo_coadd->earth_rad = 0;	
    for (i=0; i< n_coadd; i++)
    {
	geo_coadd->sat_geod_ht += geo_L[i].sat_geod_ht;
	geo_coadd->earth_rad += geo_L[i].earth_rad;
    }
    geo_coadd->sat_geod_ht /= n_coadd;
    geo_coadd->earth_rad /= n_coadd;


    n0 = 0;
    n1 = n_coadd/2 -1;
    n2 = n_coadd -1;

    /*  angles */
    geo_coadd->sol_zen_angle_toa[0] =
	geo_L[n0].sol_zen_angle_toa[0];
    geo_coadd->sol_zen_angle_toa[1] =
	geo_L[n1].sol_zen_angle_toa[2];
    geo_coadd->sol_zen_angle_toa[2] =
	geo_L[n2].sol_zen_angle_toa[2];
    
    geo_coadd->los_zen_angle_toa[0] =
	geo_L[n0].los_zen_angle_toa[0];
    geo_coadd->los_zen_angle_toa[1] =
	geo_L[n1].los_zen_angle_toa[2];
    geo_coadd->los_zen_angle_toa[2] =
	geo_L[n2].los_zen_angle_toa[2];
    
    geo_coadd->rel_azi_angle_toa[0] =
	geo_L[n0].rel_azi_angle_toa[0];
    geo_coadd->rel_azi_angle_toa[1] =
	geo_L[n1].rel_azi_angle_toa[2];
    geo_coadd->rel_azi_angle_toa[2] =
	geo_L[n2].rel_azi_angle_toa[2];

    geo_coadd->sub_sat_point =
	middle_coord (geo_L[n1].sub_sat_point,
		      geo_L[n1+1].sub_sat_point);

    /*  tangent coordinate*/
    geo_coadd->tangent_coord[0] =
	geo_L[n0].tangent_coord[0];
    geo_coadd->tangent_coord[1] =
	geo_L[n1].tangent_coord[2];
    geo_coadd->tangent_coord[2] =
	geo_L[n2].tangent_coord[2];
    
    geo_coadd->tangent_height[0] =
	geo_L[n0].tangent_height[0];
    geo_coadd->tangent_height[1] =
	geo_L[n1].tangent_height[2];
    geo_coadd->tangent_height[2] =
	geo_L[n2].tangent_height[2];
    
    return OK;
}


/* Swap start and end geolocation in Limb Scan */

SCIA_err Geo_limb_swap ( GeoL_L2_user *geo_L)
{
    float tmp;
    Coord_deg ctmp;
    
    tmp = geo_L->sol_zen_angle_toa[0];
    geo_L->sol_zen_angle_toa[0] = geo_L->sol_zen_angle_toa[2];
    geo_L->sol_zen_angle_toa[2] = tmp;
    
    tmp = geo_L->los_zen_angle_toa[0];
    geo_L->los_zen_angle_toa[0] = geo_L->los_zen_angle_toa[2];
    geo_L->los_zen_angle_toa[2] = tmp;
    
    tmp = geo_L->rel_azi_angle_toa[0];
    geo_L->rel_azi_angle_toa[0] = geo_L->rel_azi_angle_toa[2];
    geo_L->rel_azi_angle_toa[2] = tmp;

    tmp = geo_L->tangent_height[0];
    geo_L->tangent_height[0] = geo_L->tangent_height[2];
    geo_L->tangent_height[2] = tmp;
    
    ctmp = geo_L->tangent_coord[0];
    geo_L->tangent_coord[0] = geo_L->tangent_coord[2];
    geo_L->tangent_coord[2] = ctmp;
    
    return OK;
}




/* SCIA_err Geo_Limb_mean (GeoL_L2_user *geo_arr, */
/* 			int n_geo_L, */
/* 			GeoL_L2_user *geo) */
/* { */
/*     int n_first, n_last, n_mid; */
/*     int flag_even;		/\* if we have numbers of geolocations */
/* 				     mid of IT is between n_mid/n_mid+1 *\/ */
/* 				/\*  geolocation is ordered top - down *\/ */
/* 				/\*  but down is first in time *\/ */
/*     n_first = n_geo_L - 1; */
/*     n_last = 0; */
/*     n_mid = (n_geo_L -1)/2; */
/*     if (2 * n_mid != n_geo_L -1) */
/* 	flag_even = 1; */
/*     else */
/* 	flag_even = 0; */
/* 				/\*  These are take from the first ones *\/ */
/*     geo->dsr_time = geo_arr[n_first].dsr_time; */
/*     geo->attach_flag = geo_arr[n_first].attach_flag; */
/*     geo->integr_time = geo_arr[n_first].integr_time; */
/* 				/\*  for mean take always the values from */
/* 				    the mid of the integration times of */
/* 				    the individual measurements. */
/* 				    e.g. mid of swath *\/ */
/* 				/\*  first in time *\/ */
/*     geo->sol_zen_angle_toa[0] = geo_arr[n_first].sol_zen_angle_toa[1]; */
/*     geo->los_zen_angle_toa[0] = geo_arr[n_first].los_zen_angle_toa[1]; */
/*     geo->rel_azi_angle_toa[0] = geo_arr[n_first].rel_azi_angle_toa[1]; */
/*     geo->tangent_height   [0] = geo_arr[n_first].tangent_height   [1];     */
/*     geo->tangent_coord    [0] = geo_arr[n_first].tangent_coord    [1]; */
/* 				/\*  last in time*\/ */
/*     geo->sol_zen_angle_toa[2] = geo_arr[n_last].sol_zen_angle_toa[1]; */
/*     geo->los_zen_angle_toa[2] = geo_arr[n_last].los_zen_angle_toa[1]; */
/*     geo->rel_azi_angle_toa[2] = geo_arr[n_last].rel_azi_angle_toa[1]; */
/*     geo->tangent_height   [2] = geo_arr[n_last].tangent_height   [1];     */
/*     geo->tangent_coord    [2] = geo_arr[n_last].tangent_coord    [1]; */
/* 				/\*  mid of time *\/ */
/* 				/\*  even elements: between mid/mid+1*\/ */
/*     if (flag_even) */
/*     { */
/* 	geo->sol_zen_angle_toa[1] = */
/* 	    (geo_arr[n_mid].sol_zen_angle_toa[1] + */
/* 	     geo_arr[n_mid+1].sol_zen_angle_toa[1]) / 2.0; */
/* 	geo->los_zen_angle_toa[1] = */
/* 	    (geo_arr[n_mid].los_zen_angle_toa[1] + */
/* 	     geo_arr[n_mid+1].los_zen_angle_toa[1]) / 2.0; */
/* 	geo->rel_azi_angle_toa[1] = */
/* 	    (geo_arr[n_mid].rel_azi_angle_toa[1] + */
/* 	     geo_arr[n_mid].rel_azi_angle_toa[1]) / 2.0; */
/* 	geo->tangent_height   [1] = */
/* 	    (geo_arr[n_mid].tangent_height   [1] + */
/* 	     geo_arr[n_mid].tangent_height   [1]) / 2.0; */
	
/* 	geo->tangent_coord    [1] = */
/* 	    middle_coord (geo_arr[n_mid].tangent_coord    [1], */
/* 			  geo_arr[n_mid].tangent_coord    [1]); */
/*     } */
/*     else			/\* odd is easy *\/ */
/*     { */
/* 	geo->sol_zen_angle_toa[1] = geo_arr[n_mid].sol_zen_angle_toa[1]; */
/* 	geo->los_zen_angle_toa[1] = geo_arr[n_mid].los_zen_angle_toa[1]; */
/* 	geo->rel_azi_angle_toa[1] = geo_arr[n_mid].rel_azi_angle_toa[1]; */
/* 	geo->tangent_height   [1] = geo_arr[n_mid].tangent_height   [1];     */
/* 	geo->tangent_coord    [1] = geo_arr[n_mid].tangent_coord    [1]; */
/*     } */
/*     return OK; */
/* } */

/* Second approach for mean Limb geolocation:
   choose a represantitive height LIMB_MEAN_HEIGHT_IDX = 27 =~ 20km for
   angles and location */

int limb_mean_height_idx=27;

SCIA_err Geo_Limb_mean (GeoL_L2_user *geo_arr,
			GeoL_L2_user *geo,
			int n_level)
{
    int n;
    int n_first;

    n_first = n_level - 1;
    
    /*  These are take from the first ones */
    geo->dsr_time = geo_arr[n_first].dsr_time;
    geo->attach_flag = geo_arr[n_first].attach_flag;
    geo->integr_time = geo_arr[n_first].integr_time;
    /*  for mean chose values from
	LIMB_MEAN_HEIGHT_IDX e.g. about 20km
	first in time */
    for (n=0; n<3; n++)
    {
	geo->sol_zen_angle_toa[n] =
	    geo_arr[limb_mean_height_idx].sol_zen_angle_toa[n];
	geo->los_zen_angle_toa[n] =
	    geo_arr[limb_mean_height_idx].los_zen_angle_toa[n];
	geo->rel_azi_angle_toa[n] =
	    geo_arr[limb_mean_height_idx].rel_azi_angle_toa[n];
	geo->tangent_height   [n] =
	    geo_arr[limb_mean_height_idx].tangent_height   [n];    
	geo->tangent_coord    [n] =
	    geo_arr[limb_mean_height_idx].tangent_coord    [n];
    }

    geo->sat_geod_ht   = geo_arr[limb_mean_height_idx].sat_geod_ht;
    geo->earth_rad     = geo_arr[limb_mean_height_idx].earth_rad;
    geo->sub_sat_point = geo_arr[limb_mean_height_idx].sub_sat_point;

    return OK;
}    





Coord_deg middle_coord (Coord_deg p1, Coord_deg p2)
{
    Coord_deg result;
    /*  lat is simple */
    result.lat = (p1.lat+p2.lat)/2;
    /*  case 1 for lon */
    if (p1.lon < -90.0 && p2.lon > 90.0)
    {
	p1.lon += 360;
	result.lon = (p1.lon+p2.lon)/2;
	if ( result.lon > 180.0)
	    result.lon -= 360.0;
	return result;
    }
    /*  case 2 for lon */
    if (p2.lon < -90.0 && p1.lon > 90.0)
    {
	p2.lon += 360;
	result.lon = (p1.lon+p2.lon)/2;
	if ( result.lon > 180.0)
	    result.lon -= 360.0;
	return result;
    }
    /*  default */
    result.lon = (p1.lon+p2.lon)/2;
    return result;
}


char mipas_species [N_GAS_MAX][30] = {
"H2O",
"O3",
"HNO3",
"CH4",
"N2O",
"NO2"};


char Generic_Prod [MAX_GENERIC_PROD][15] = {
    "CLOUDS_AEROSOL",
    "O3",
    "NO2",
    "O3vis",
    "BRO",
    "OCLO",
    "SO2",
    "HCHO",
    "H2O",
    "CO2",
    "N2O",
    "CH4",
    "CO",
    "HNO3"
};

/* table for allocation of real products to generic products */
/* has to be adjusted, when fittingwindows are fixed/known for NRT/OL */

int gen_product_matrix [MAX_GENERIC_PROD][6] = {
    /* NRT                        OL                         MET        LIMB   GOME     MIPAS*/
    /*CLOUDS_AEROSOL, */ {SCI_NL__2P_CLOUDS_AEROSOL, SCI_OL__2P_CLOUDS_AEROSOL, SCI_RV__2P_MAX_DS_NAME,         SCI_OL__2P_MAX_DS_NAME    ,GOME_L2_MOL_O3 , MIPAS_EMPTY},
    /*O3,             */ {SCI_NL__2P_DOAS_0_O3,      SCI_OL__2P_NAD_UV0_O3,     SCI_RV__2P_O3_VERTICAL_COLUMN,  SCI_OL__2P_LIM_UV0_O3     ,GOME_L2_MOL_O3 , MIPAS_O3   },
    /*NO2,            */ {SCI_NL__2P_DOAS_1_NO2,     SCI_OL__2P_NAD_UV1_NO2,    SCI_RV__2P_MAX_DS_NAME, 	SCI_OL__2P_LIM_UV1_NO2    ,GOME_L2_MOL_NO2, MIPAS_NO2  },
    /*O3vis,          */ {SCI_NL__2P_DOAS_1_O3,      SCI_OL__2P_NAD_UV2_O3,     SCI_RV__2P_MAX_DS_NAME,         SCI_OL__2P_LIM_UV2_O3     ,GOME_L2_MAX    , MIPAS_EMPTY},
    /*BRO,            */ {SCI_NL__2P_DOAS_2_BRO,     SCI_OL__2P_NAD_UV3_BRO,    SCI_RV__2P_MAX_DS_NAME, 	SCI_OL__2P_LIM_UV3_BRO    ,GOME_L2_MAX    , MIPAS_EMPTY},
    /*OCLO,           */ {SCI_NL__2P_DOAS_2_OCLO,    SCI_OL__2P_NAD_UV6_OCLO,   SCI_RV__2P_MAX_DS_NAME,  	SCI_OL__2P_LIM_UV6_OCLO   ,GOME_L2_MAX    , MIPAS_EMPTY},
    /*SO2,            */ {SCI_NL__2P_DOAS_4_SO2,     SCI_OL__2P_NAD_UV5_SO2,    SCI_RV__2P_MAX_DS_NAME, 	SCI_OL__2P_LIM_UV5_SO2    ,GOME_L2_MAX    , MIPAS_EMPTY},
    /*HCHO,           */ {SCI_NL__2P_DOAS_5_HCHO,    SCI_OL__2P_NAD_UV4_H2CO,   SCI_RV__2P_MAX_DS_NAME, 	SCI_OL__2P_LIM_UV4_H2CO   ,GOME_L2_MAX    , MIPAS_EMPTY},
    /*H20,            */ {SCI_NL__2P_BIAS_0_H2O,     SCI_OL__2P_NAD_IR0_H2O,    SCI_RV__2P_MAX_DS_NAME, 	SCI_OL__2P_LIM_IR0_H2O    ,GOME_L2_MAX    , MIPAS_H2O  },
    /*CO2,            */ {SCI_NL__2P_BIAS_0_CO2,     SCI_OL__2P_NAD_IR4_CO2,    SCI_RV__2P_MAX_DS_NAME, 	SCI_OL__2P_MAX_DS_NAME    ,GOME_L2_MAX    , MIPAS_EMPTY},
    /*N20,            */ {SCI_NL__2P_BIAS_1_N2O,     SCI_OL__2P_NAD_IR2_N2O,    SCI_RV__2P_MAX_DS_NAME, 	SCI_OL__2P_LIM_IR2_N2O    ,GOME_L2_MAX    , MIPAS_N2O  },
    /*CH4,            */ {SCI_NL__2P_BIAS_1_CH4,     SCI_OL__2P_NAD_IR1_CH4,    SCI_RV__2P_MAX_DS_NAME,         SCI_OL__2P_LIM_IR1_CH4    ,GOME_L2_MAX    , MIPAS_CH4  },
    /*CO,             */ {SCI_NL__2P_BIAS_2_CO,      SCI_OL__2P_NAD_IR3_CO,     SCI_RV__2P_MAX_DS_NAME,         SCI_OL__2P_LIM_IR3_CO     ,GOME_L2_MAX    , MIPAS_EMPTY},
    /*HNO3,           */ {SCI_NL__2P_MAX_DS_NAME,    SCI_OL__2P_MAX_DS_NAME,    SCI_RV__2P_MAX_DS_NAME,         SCI_OL__2P_MAX_DS_NAME    ,GOME_L2_MAX    , MIPAS_HNO3 }
} ;


product_id prod_id_str (const char* prod_str)
{
    product_id prod;
    
    int i;
    /* Search for generic Name */
    for (i = GEN_CLOUDS_AEROSOL; i<=MAX_GENERIC_PROD; i++)
    {
	if (strcmp (prod_str, Generic_Prod[i]) == 0)
	{
	    prod.prod_n = gen_product_matrix[i][0];
	    prod.prod_o = gen_product_matrix[i][1];
	    prod.prod_m = gen_product_matrix[i][2];
	    prod.prod_l = gen_product_matrix[i][3];
	    prod.prod_g = gen_product_matrix[i][4];
	    prod.prod_mip = gen_product_matrix[i][5];
	    return prod;		/* generic name identified! */	    
	}
    }
    /*  pre-set all to not available */
    prod.prod_n = SCI_NL__2P_MAX_DS_NAME;
    prod.prod_o = SCI_OL__2P_MAX_DS_NAME;
    prod.prod_m = SCI_RV__2P_MAX_DS_NAME;
    prod.prod_l = SCI_OL__2P_MAX_DS_NAME;
    prod.prod_mip = MIPAS_EMPTY;
    prod.prod_g = GOME_L2_MAX;


    /* Search for NRT-Name */
    for (i = SCI_NL__2P_CLOUDS_AEROSOL; i <= SCI_NL__2P_BIAS_SPARE_2; i++)
    {
	if (strcmp (prod_str, DS_NAME_SCI_NL__2P[i]+5) == 0)
	{			/* NRT Name identified */
	    prod.prod_n = i;
	}
    }
    /* Search for OL-Name */
    for (i = SCI_OL__2P_CLOUDS_AEROSOL; i <= SCI_OL__2P_NAD_IR5_SPARE; i++)
    {
	if (strcmp (prod_str, DS_NAME_SCI_OL__2P[i]+4) == 0)
	{			/* OL nadir identified */
	    prod.prod_o = i;
	}
    }
    
    for (i = SCI_OL__2P_LIM_UV0_O3; i <= SCI_OL__2P_LIM_IR4_SPARE; i++)
    {
	if (strcmp (prod_str, DS_NAME_SCI_OL__2P[i]+4) == 0)
	{			/* OL limb identified */
	    prod.prod_l = i;
	}
    }

    /*  Meteo O3 requested? */
    if (strcmp (prod_str, DS_NAME_SCI_RV__2P[SCI_RV__2P_O3_VERTICAL_COLUMN]) == 0)
    {
	prod.prod_m = SCI_RV__2P_O3_VERTICAL_COLUMN;
    }
    
    /* GOME product requested */
    for (i = GOME_L2_MOL_O3; i < GOME_L2_MAX; i++)
    {
	if (strcmp (prod_str, GOME_L2_MOL_NAME[i]) == 0)
	{
	    prod.prod_g = i;
	}
    }
    /* MIPAS products all have the generic name  */
	    
    /*  return what is found */
    return prod;          
}


/* Print out available products */
void prod_id_print (FILE *out)
{
    int n,i;
    /*  List all the names for Nadir */
    fprintf (out,"   \n"
	     "List of available product Identifikationstrings:\n"
	     "================================================\n"
	     "\n"
	     "-- Nadir:\n");
    fprintf (out, " - NRT: UV/vis (DOAS):\n");
    for (n=0, i = SCI_NL__2P_DOAS_0_O3; i <= SCI_NL__2P_DOAS_5_O4; i++, n++)
    {
	if ( n % 7 == 0 )
	    fprintf (out, "      **");
 	fprintf (out, " %-7s", DS_NAME_SCI_NL__2P[i]+5);
	if ( n % 7 == 6 )
	    fprintf (out, "\n");
    }
    fprintf (out, "\n\n - NRT: IR (BIAS):\n");
    for (n=0, i = SCI_NL__2P_BIAS_0_H2O; i <= SCI_NL__2P_BIAS_2_CH4; i++, n++)
    {
	if ( n % 7 == 0 )
	    fprintf (out, "      **");
	fprintf (out, " %-7s", DS_NAME_SCI_NL__2P[i]+5);
	if ( n % 7 == 6 )
	    fprintf (out, "\n");
    }
    fprintf (out, "\n - NRT: Clouds and Aerosol:\n");
    fprintf (out, "      ** %s",
	     DS_NAME_SCI_NL__2P[SCI_NL__2P_CLOUDS_AEROSOL]);
	     
    fprintf (out, "\n - OL: UV/vis (DOAS):\n");
    for (n=0, i = SCI_OL__2P_NAD_UV0_O3; i <= SCI_OL__2P_NAD_UV6_OCLO; i++, n++)
    {
	if ( n % 7 == 0 )
	    fprintf (out, "      **");
	fprintf (out, " %-7s", DS_NAME_SCI_OL__2P[i]+4);
	if ( n % 7 == 7 )
	    fprintf (out, "\n");
    }
    fprintf (out, "\n - OL: IR (BIAS):\n");
    for (n=0, i = SCI_OL__2P_NAD_IR0_H2O; i <= SCI_OL__2P_NAD_IR4_CO2; i++, n++)
    {
	if ( n % 7 == 0 )
	    fprintf (out, "      **");
	fprintf (out, " %-7s", DS_NAME_SCI_OL__2P[i]+4);
	if ( n % 7 == 6 )
	    fprintf (out, "\n");
    }
    fprintf (out, "\n - OL: Clouds and Aerosol:\n");
    fprintf (out, "      ** %s ",
	     DS_NAME_SCI_OL__2P[SCI_OL__2P_CLOUDS_AEROSOL]);
    fprintf (out, "\n - Meteo:\n");
    fprintf (out, "      ** %-7s ",
	     DS_NAME_SCI_RV__2P[SCI_RV__2P_O3_VERTICAL_COLUMN]);
    
    /* List of product IDs for gome */
    fprintf (out, "\n - GOME:\n");
    fprintf (out, "      **");
    for (n = GOME_L2_MOL_O3; n <= GOME_L2_MAX; n++)
    {
	fprintf (out, " %-7s", GOME_L2_MOL_NAME[n]);
    }
    
    /*  List all the names for Limb */
    fprintf (out, "\n\n-- LIMB:\n");
    
    fprintf (out, " - OL Limb: UVvis:\n");
    for (n=0, i = SCI_OL__2P_LIM_UV0_O3; i <= SCI_OL__2P_LIM_UV6_OCLO; i++, n++)
    {
	if ( n % 7 == 0 )
	    fprintf (out, "      **");
	fprintf (out, " %-7s", DS_NAME_SCI_OL__2P[i]+4);
	if ( n % 7 == 6 )
	    fprintf (out, "\n");
    }
	
    fprintf (out, " - OL Limb: IR:\n");
    for (n=0, i = SCI_OL__2P_LIM_IR0_H2O; i <= SCI_OL__2P_LIM_IR3_CO; i++, n++)
    {
	if ( n % 7 == 0 )
	    fprintf (out, "      **");
	fprintf (out, " %-7s", DS_NAME_SCI_OL__2P[i]+4);
	if ( n % 7 == 7 )
	    fprintf (out, "\n");
    }
    fprintf (out, "\n - OL Limb: Pressure, Temperature, Height:\n");
    fprintf (out, "      **%-7s\n",
	     DS_NAME_SCI_OL__2P[SCI_OL__2P_LIM_PTH]+4);

    /* List for MIPAS */
    fprintf (out, "\n - MIPAS:\n");
    fprintf (out, "      **");
    for (n = 0; n <= MIPAS_EMPTY; n++)
    {
	fprintf (out, " %-7s", mipas_species[n]);
    }

    fprintf (out, "\n");
    
    /*  Write the generic names */
    fprintf (out, "A set of generic names for the most common species is defined to have\n"
	     "consistent access to all products. The following table shows the mapping\n"
	     "to the definitions in the product.\n"
	     "!!! This table might change in future versions. !!!\n");
    fprintf (out, "\n\nGeneric names:\n"
	     "-------------\n");
    fprintf (out, "                NRT             OL              OL-Limb    GOME   MIPAS Meteo\n"
	     ".............................................................................\n");
    
    for (i=0; i<MAX_GENERIC_PROD; i++)
    {
	switch (i)
	{
	    case GEN_O3:
		fprintf (out, "%-14s: %-14s  %-14s  %-9s  %-5s  %-5s %14s\n",
			 Generic_Prod[i],
			 DS_NAME_SCI_NL__2P [gen_product_matrix[i][0]]+5,
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][1]]+4,
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][3]]+4,
			 GOME_L2_MOL_NAME[gen_product_matrix[i][4]],
			 mipas_species[gen_product_matrix[i][5]],
			 DS_NAME_SCI_RV__2P [gen_product_matrix[i][2]]);
		break;
	    case GEN_NO2:
		fprintf (out, "%-14s: %-14s  %-14s  %-9s  %-5s  %-5s ---\n",
			 Generic_Prod[i],
			 DS_NAME_SCI_NL__2P [gen_product_matrix[i][0]]+5,
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][1]]+4,
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][3]]+4,
			 GOME_L2_MOL_NAME[gen_product_matrix[i][4]],
			 mipas_species[gen_product_matrix[i][5]]);
		break;
	    case GEN_CLOUDS_AEROSOL:
		fprintf (out, "%-14s: %-14s  %-14s  %-9s  %-5s  ---   ---\n",
			 Generic_Prod[i],
			 DS_NAME_SCI_NL__2P [gen_product_matrix[i][0]],
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][1]],
			 "---",
			 GOME_L2_MOL_NAME[gen_product_matrix[i][4]]);
		break;
	    case GEN_CO2:
		fprintf (out, "%-14s: %-14s  %-14s  %-9s  ---    ---   ---\n",
			 Generic_Prod[i],
			 DS_NAME_SCI_NL__2P [gen_product_matrix[i][0]]+5,
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][1]]+4,
			 "---");
		break;
	    case GEN_N2O:
	    case GEN_CH4:
	    case GEN_H2O:
 		fprintf (out, "%-14s: %-14s  %-14s  %-9s  ---    %-5s ---\n",
			 Generic_Prod[i],
			 DS_NAME_SCI_NL__2P [gen_product_matrix[i][0]]+5,
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][1]]+4,
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][3]]+4,
			 mipas_species[gen_product_matrix[i][5]]);
		break;
	    case GEN_HNO3:
 		fprintf (out, "%-14s  "
			 "---             "
			 "---             "
			 "---        "
			 "---    "
			 "%-5s ---\n",
			 Generic_Prod[i],
			 mipas_species[gen_product_matrix[i][5]]);
		break; 

	    default:
 		fprintf (out, "%-14s: %-14s  %-14s  %-9s  ---    ---   ---\n",
			 Generic_Prod[i],
			 DS_NAME_SCI_NL__2P [gen_product_matrix[i][0]]+5,
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][1]]+4,
			 DS_NAME_SCI_OL__2P [gen_product_matrix[i][3]]+4);
		break;
	}
    }

    

}
	     
    

/* read Geolocatint Level 2 NRT */
void GeoN_L2_user_array_getbin (FILE* unit, GeoN_L2_user *var, int nr)
{
    int i,n;
    Coord cor_coord[4];
    Coord cen_coord;
    int sum_it = 0;
    
    for (n=0; n<nr; n++)
    {
	MJD_getbin (unit, &var[n].dsr_time);
	char_getbin (unit, &var[n].attach_flag);
	ushort_getbin (unit, &var[n].integr_time);
	
	float_array_getbin (unit, var[n].sol_zen_angle_toa, 3);
	float_array_getbin (unit, var[n].los_zen_angle_toa, 3);
	float_array_getbin (unit, var[n].rel_azi_angle_toa, 3);
	float_getbin (unit, &var[n].sat_geod_ht);
	float_getbin (unit, &var[n].earth_rad);
		
	var[n].sub_sat_point.lat = 0.0;
	var[n].sub_sat_point.lon = 0.0;
	
	Coord_array_getbin (unit, cor_coord, 4);
	for (i=0; i<4; i++)
	{
	    var[n].cor_coor_nad[i].lat = cor_coord[i].lat *  ANGLE_FAC;
	    var[n].cor_coor_nad[i].lon = cor_coord[i].lon *  ANGLE_FAC;
	}
	Coord_getbin (unit, &cen_coord);
	var[n].cen_coor_nad.lat = cen_coord.lat *  ANGLE_FAC;
	var[n].cen_coor_nad.lon = cen_coord.lon *  ANGLE_FAC;
	/* determine subset_counter */
	/* Large Pixels, usual 10s : -2 */
	if ( var[n].integr_time >= 80 )
	    var[n].subset_counter = -1 * var[n].integr_time % 80;
	else
	    /*  0 - 63 forward */
	    /*  64 - 79 backward */
	{
	    var[n].subset_counter = sum_it;
	    sum_it = (sum_it + var[n].integr_time) % 80;
	}
	    
    }
}

/* read Geolocation Level 2 OL */

void GeoN_L2_OL_user_array_getbin (FILE* unit, GeoN_L2_user *var, int nr)
{
    int i,n;
    Coord cor_coord[4];
    Coord cen_coord;
    Coord sub_sat_point;
    int sum_it = 0;
    
    for (n=0; n<nr; n++)
    {
	MJD_getbin (unit, &var[n].dsr_time);
	char_getbin (unit, &var[n].attach_flag);
	ushort_getbin (unit, &var[n].integr_time);
	
	float_array_getbin (unit, var[n].sol_zen_angle_toa, 3);
	float_array_getbin (unit, var[n].los_zen_angle_toa, 3);
	float_array_getbin (unit, var[n].rel_azi_angle_toa, 3);
	float_getbin (unit, &var[n].sat_geod_ht);
	float_getbin (unit, &var[n].earth_rad);

	Coord_getbin (unit, &sub_sat_point);
	var[n].sub_sat_point.lat = cen_coord.lat *  ANGLE_FAC;
	var[n].sub_sat_point.lon = cen_coord.lon *  ANGLE_FAC;

	Coord_array_getbin (unit, cor_coord, 4);
	for (i=0; i<4; i++)
	{
	    var[n].cor_coor_nad[i].lat = cor_coord[i].lat *  ANGLE_FAC;
	    var[n].cor_coor_nad[i].lon = cor_coord[i].lon *  ANGLE_FAC;
	}
	Coord_getbin (unit, &cen_coord);
	var[n].cen_coor_nad.lat = cen_coord.lat *  ANGLE_FAC;
	var[n].cen_coor_nad.lon = cen_coord.lon *  ANGLE_FAC;
	/* determine subset_counter */
	/* Large Pixels, usual 10s : -2 */
	if ( var[n].integr_time >= 80 )
	    var[n].subset_counter = -1 * var[n].integr_time % 80;
	else
	    /*  0 - 63 forward */
	    /*  64 - 79 backward */
	{
	    var[n].subset_counter = sum_it;
	    sum_it = (sum_it + var[n].integr_time) % 80;
	}
	    
    }
}

/* read Geolocation Level 2 OL - Limb */
/* In contrast to the wording in the IODD (Doc.No.: ENV-ID-DLR-SCI-2200-4
   Issue: 4/B Date: 27. Aug 2002) and the definitions in EnviView, most
   angles and heights in OL are stored as float numbers, not as integers
   with a factor to multiply....
   We will see, what is in the final product...
*/


void GeoL_L2_OL_user_array_getbin (FILE* unit, GeoL_L2_user *var, int nr)
{
    int i,n;
/*    int  angle[3]; */
    Coord tangent_coord[3];
    Coord sub_sat_point;
/*    int height; */
    
    for (n=0; n<nr; n++)
    {
	MJD_getbin (unit, &var[n].dsr_time);
	char_getbin (unit, &var[n].attach_flag);
	ushort_getbin (unit, &var[n].integr_time);

/*	int_array_getbin (unit, angle, 3);
	for (i=0; i<3; i++)
	var[n].sol_zen_angle_toa[i] = angle[i] *  ANGLE_FAC;	
	int_array_getbin (unit, angle, 3);
	for (i=0; i<3; i++)
	var[n].los_zen_angle_toa[i] = angle[i] *  ANGLE_FAC;	
	int_array_getbin (unit, angle, 3);
	for (i=0; i<3; i++)
	var[n].rel_azi_angle_toa[i] = angle[i] *  ANGLE_FAC; */
	float_array_getbin (unit, var[n].sol_zen_angle_toa, 3);
	float_array_getbin (unit, var[n].los_zen_angle_toa, 3);
	float_array_getbin (unit, var[n].rel_azi_angle_toa, 3);

	
/*	uint_getbin (unit, &height);
	var[n].sat_geod_ht = height * HEIGHT_FAC;
	uint_getbin (unit, &height);
	var[n].earth_rad = height * HEIGHT_FAC; */
	float_getbin (unit, &var[n].sat_geod_ht );
	float_getbin (unit, &var[n].earth_rad );
	
	Coord_getbin (unit, &sub_sat_point);
	var[n].sub_sat_point.lat = sub_sat_point.lat *  ANGLE_FAC;
	var[n].sub_sat_point.lon = sub_sat_point.lon *  ANGLE_FAC;

	Coord_array_getbin (unit, tangent_coord, 3);
	for (i=0; i<3; i++)
	{
	    var[n].tangent_coord[i].lat = tangent_coord[i].lat *  ANGLE_FAC;
	    var[n].tangent_coord[i].lon = tangent_coord[i].lon *  ANGLE_FAC;
	}

/*	
  int_array_getbin (unit, angle, 3);
  for (i=0; i<3; i++)
  var[n].tangent_height[i] = angle[i] *  HEIGHT_FAC; */
	
	float_array_getbin (unit, var[n].tangent_height, 3);

    }
}



/* read Geolocatint Level 2 Met */

void GeoN_L2_user_met_array_getbin (FILE* unit, GeoN_L2_user *var, int nr)
{
    int i,n;
    Coord cor_coord[4];
    int sum_it = 0;
    
    for (n=0; n<nr; n++)
    {
	MJD_getbin (unit, &var[n].dsr_time);
	char_getbin (unit, &var[n].attach_flag);
	ushort_getbin (unit, &var[n].integr_time);
	/* zero unused values */
	for (i=0; i<3; i++)
	{
	    var[n].sol_zen_angle_toa[i] = 0.0;
	    var[n].los_zen_angle_toa[i] = 0.0;
	    var[n].rel_azi_angle_toa[i] = 0.0;
	}
	var[n].sat_geod_ht = 0.0;
	var[n].earth_rad = 0.0;
	
	Coord_array_getbin (unit, cor_coord, 4);
	for (i=0; i<4; i++)
	{
	    var[n].cor_coor_nad[i].lat = cor_coord[i].lat *  ANGLE_FAC;
	    var[n].cor_coor_nad[i].lon = cor_coord[i].lon *  ANGLE_FAC;
	}
	/*  calculate center middle of point 1 and 4 */
	var[n].cen_coor_nad = middle_coord (var[n].cor_coor_nad[0],
					    var[n].cor_coor_nad[3]);
	/* determine subset_counter */
	/* determine subset_counter */
	/* Large Pixels, usual 10s : -2 */
	if ( var[n].integr_time >= 80 )
	    var[n].subset_counter = -1 * var[n].integr_time % 80;
	else
	    /*  0 - 63 forward */
	    /*  64 - 79 backward */
	{
	    var[n].subset_counter = sum_it;
	    sum_it = (sum_it + var[n].integr_time) % 80;
	}

    }
}





SCIA_err L2N_info (FILE* out, info_l2N* info)

{
    char fitwin_id_str[2];
    int i,fitwin_id,n;
    fprintf (out, "--------------------------------------------------------------\n");

    switch (info->scia_product)
    {
	case SCIA_L2N:
	case SCIA_L2O:
	case SCIA_L2M:
	    fprintf (out, "%s\n", info->mph.product);
	    fprintf (out, "Processing software version: %s\n", info->mph.software_ver);
	    fprintf (out, "Processing time: %s\n", info->mph.proc_time);
	    fprintf (out, "Sensing start: %s\n", info->mph.sensing_start);
	    fprintf (out, "Sensing stop : %s\n", info->mph.sensing_stop);
	    fprintf (out, "Orbit number : %s\n", info->mph.abs_orbit);
	    break;
	case GOME_L2:
	    fprintf (out, "%s\n", info->mph_gome.pir);
	    fprintf (out, "Processing software version: GDP %s %s Format %s\n", 
		     info->sph_gome.version_DOAS,
		     info->sph_gome.version_keyparam,
		     info->sph_gome.version_format);
	    fprintf (out, "Processing time: %2.2s-%2.2s-%4.4s %2.2s:%2.2s:%2.2s\n", 
		     info->mph_gome.pir+30,
		     info->mph_gome.pir+28,
		     info->mph_gome.pir+24,
		     info->mph_gome.pir+32,
		     info->mph_gome.pir+34,
		     info->mph_gome.pir+36);
	    fprintf (out, "Orbit number : %5.5s\n", info->mph_gome.pir+6);
	    break;
	default:
	    return ERROR;
	    break;
    }

    switch (info->scia_product)
    {
    /*  NRT product */
	case SCIA_L2N:

	fprintf (out, "---------------------------------\n");
	fprintf (out, "CLOUDS_AEROSOL : %6d DSR\n",
		 info->dsd[SCI_NL__2P_CLOUDS_AEROSOL].num_dsr);
	fprintf (out, "---------------------------------\n"
		 "TYPE  Product  #DSR    Fitwindows\n");
	for ( i=SCI_NL__2P_DOAS_0_O3; i<=SCI_NL__2P_BIAS_SPARE_2; i++ )
	{
	    if ( info->dsd[i].num_dsr == 0 )
		continue;      
	    fitwin_id_str[0] = DS_NAME_SCI_NL__2P[i][5];
	    fitwin_id_str[1] = '\0';
	    fitwin_id = atoi (fitwin_id_str);

	    if ( i<= SCI_NL__2P_DOAS_SPARE_2 )
		fprintf (out, "DOAS  %7s  %6d  %.9s\n",
			 DS_NAME_SCI_NL__2P[i]+5, info->dsd[i].num_dsr,
			 info->sph.doas_fitting_window[ fitwin_id ]);
	    else
		fprintf (out, "BIAS  %7s  %6d  %.44s\n",
			 DS_NAME_SCI_NL__2P[i]+5, info->dsd[i].num_dsr,
			 info->sph.bias_fitting_window[ fitwin_id ]);
	}
	
	fprintf (out, "---------------------------------\n"
		 "Unused DOAS: ");
	for ( n=0,i=SCI_NL__2P_DOAS_0_O3; i<=SCI_NL__2P_DOAS_SPARE_2; i++ )
	{
	    if ( info->dsd[i].num_dsr != 0 )
		continue;
	    if (n++ == 7)
	    {
		fprintf (out, "\n             ");
		n = 1;
	    }
	    fprintf (out, "%s ",DS_NAME_SCI_NL__2P[i]+5);

	}
	fprintf (out, "\nUnused BIAS: ");
	for (n=0, i=SCI_NL__2P_BIAS_0_H2O; i<=SCI_NL__2P_BIAS_SPARE_2; i++ )
	{
	    if ( info->dsd[i].num_dsr != 0 )
		continue;
	    if (n++ == 7)
	    {
		fprintf (out, "\n             ");
		n = 1;
	    }
	    fprintf (out, "%s ",DS_NAME_SCI_NL__2P[i]+5);
	}
	fprintf (out, "\n---------------------------------\n");
	return OK;
	break;


    /*  Offline Products*/
	case SCIA_L2O:

	fprintf (out, "---------------------------------\n");
	fprintf (out, "CLOUDS_AEROSOL : %6d DSR\n",
		 info->dsd[SCI_OL__2P_CLOUDS_AEROSOL].num_dsr);
	fprintf (out, "---------------------------------\n"
		 "TYPE        Product    #DSR   Fitwindows\n");
	/* Nadir */
	for ( i=SCI_OL__2P_NAD_UV0_O3; i<=SCI_OL__2P_NAD_IR5_SPARE; i++ )
	{
	    if ( info->dsd[i].num_dsr == 0 )
		continue;      
	    fitwin_id_str[0] = DS_NAME_SCI_OL__2P[i][6];
	    fitwin_id_str[1] = '\0';
	    fitwin_id = atoi (fitwin_id_str);
	    
	    if ( i<= SCI_OL__2P_NAD_UV7_SPARE )
		fprintf (out, "NADIR UV  %9s  %6d  %.30s\n",
			 DS_NAME_SCI_OL__2P[i]+4, info->dsd[i].num_dsr,
			 info->sph_ol.nad_fit_window_uv[ fitwin_id ]);
	    else
		fprintf (out, "NADIR IR    %7s  %6d  %.30s\n",
			 DS_NAME_SCI_OL__2P[i]+4, info->dsd[i].num_dsr,
			 info->sph_ol.nad_fit_window_ir[ fitwin_id ]);
	}

	/* Limb */
	i = SCI_OL__2P_LIM_PTH;
	if ( info->dsd[i].num_dsr > 0 )
		fprintf (out, "LIM PTH   p T H    %6d   %.15s\n",
			 info->dsd[i].num_dsr,
			 info->sph_ol.lim_fit_window_pth);

	for ( i=SCI_OL__2P_LIM_UV0_O3; i<=SCI_OL__2P_LIM_IR4_SPARE; i++ )
	{
	    if ( info->dsd[i].num_dsr == 0 )
		continue;      
	    fitwin_id_str[0] = DS_NAME_SCI_OL__2P[i][6];
	    fitwin_id_str[1] = '\0';
	    fitwin_id = atoi (fitwin_id_str);
	    
	    if ( i<= SCI_OL__2P_LIM_UV7_SPARE )
		fprintf (out, "LIMB UV   %9s  %6d  %.15s\n",
			 DS_NAME_SCI_OL__2P[i]+4, info->dsd[i].num_dsr,
			 info->sph_ol.lim_fit_window_uv[ fitwin_id ]);
	    else
		fprintf (out, "LIMB IR   %7s  %6d  %.15s\n",
			 DS_NAME_SCI_OL__2P[i]+4, info->dsd[i].num_dsr,
			 info->sph_ol.lim_fit_window_ir[ fitwin_id ]);
	}

	
	fprintf (out, "---------------------------------\n"
		 "Unused: ");
	for ( n=0,i=SCI_OL__2P_NAD_UV0_O3; i<=SCI_OL__2P_LIM_IR4_SPARE; i++ )
	{
	    if ( info->dsd[i].num_dsr != 0 )
		continue;
	    if (n++ == 5)
	    {
		fprintf (out, "\n        ");
		n = 1;
	    }
	    fprintf (out, "%s ",DS_NAME_SCI_OL__2P[i]);

	}
	fprintf (out, "\n---------------------------------\n");
	return OK;
	break;
    /*  Meteo products */
	case SCIA_L2M:
	fprintf (out, "---------------------------------\n");
	fprintf (out, "O3_VERTICAL_COLUMN : %6d DSR\n",
		 info->dsd[SCI_RV__2P_O3_VERTICAL_COLUMN].num_dsr);
	fprintf (out, "---------------------------------\n");
	return OK;
	break;
	
	/* GOME products */
	case GOME_L2:
	    fprintf (out, "---------------------------------\n");
	    fprintf (out, "Number of DOAS data records: %4i\n",
		     info->mph_gome.n_doas);
	    for (n=0; n<info->sph_gome.Nmol; n++)
	    {
		i = atoi(info->sph_gome.molpair[n])-1;
		if (i < 0 || i > info->sph_gome.Nmol-1)
		    return ERROR;
		fprintf (out, "%6.6s  %6.1f-%6.1f\n",
			 info->sph_gome.molpair[n],
			 info->sph_gome.windowpair[i][0],
			 info->sph_gome.windowpair[i][1]);	
	    }
	    fprintf (out, "---------------------------------\n");
	    break;
	default:
	    return ERROR;
	    break;	    
    }
    return ERROR;
}	 

/* Write the initialisation file (XML) of an OL produkt */

SCIA_err write_OL_initialisation_file (char *xml_dir, info_l2N *info)

{
    FILE* out;

    char out_name[2001];
    char *xml;
    int size, size_read;
    /*  check OL product and inititalisation file
	available*/
    if ( info->scia_product != SCIA_L2O  ||
	 info->dsd[ SCI_OL__2P_STATIC_PARAM ].num_dsr != 1 )
    {
	return ERROR;
    }
    /*  goto Ini file in product */
    if (set_DSD_offset( info->FILE_l2N,
			&info->dsd[ SCI_OL__2P_STATIC_PARAM ] )
	!= OK)
    {
	return ERROR;
    }
    /*  setup output file name, use original
	Initialisation filename */
    sprintf (out_name, "%s/%s.xml", xml_dir,
	     info->dsd[ SCI_OL__2P_INITIALISATION_FILE ].filename);
    /*  open file */
    if ( (out = fopen(out_name, "w")) == NULL)
	return ERROR;

    size=info->dsd[ SCI_OL__2P_STATIC_PARAM ].size;
    /*  make storing area in memory */
    if ( (xml = malloc (size)) == NULL)
	return ERROR;
    /*  read and check */
    size_read = fread (xml, sizeof(char), size, info->FILE_l2N);

    if (size_read != size)
	return ERROR;
    /*  write and check */
    size_read = fwrite (xml, sizeof(char), size, out);
    
    if (size_read != size)
	return ERROR;
    /*  close file, free memory */
    free (xml);
    fclose (out);
    
    return OK;
}

	
/* Check, if file is a scia or a gome product */
/* return value:
   1 : true : seems to be a GOME Product
   0 : false : not a GOME product
*/

int gome_check (FILE* file)
{
    /*  first 5 Letter in GOME Lv2 */
    char GOME_ident[6] = "E2GOM";
    char ident_str[6];
    /*  read 5 letters */
    char_array_getbin (file, ident_str, 5);
    ident_str[5] = '\0';
    /*  move back to zero position */
    fseek (file, 0, SEEK_SET);
    /*  compare read chars with Identifier */
    if (strcmp (GOME_ident, ident_str) == 0)
	return 1;		/* GOME */
    else
	return 0;		/* no GOME */
}
    


/* Calculate distance between center coordinate and point p */

double distance (user_data *ud, Coord_deg p)
{
    double earth_rad = ud->geo.earth_rad;
    if (earth_rad == 0.0)
	earth_rad = 6386.0;	/* Meteo products have no earth radius */
    return earth_rad *
	circle_distance (ud->geo.cen_coor_nad, p);
}

/* Calculate distance between center coordinate and point p (Limb)*/

double distance_limb (user_data *ud, Coord_deg p)
{
    double earth_rad = ud->geo_L.earth_rad;
    if (earth_rad == 0.0)
	earth_rad = 6386.0;	/* Meteo products have no earth radius */
    return earth_rad *
	circle_distance (ud->geo_L.tangent_coord[1], p);
}


